/*
 *  File:        math_simbols.C
 *  Purpose:     User interface to math symbols
 *  Author:      Alejandro Aguilar Sierra <asierra@servidor.unam.mx>
 *  Created:     November 1995
 *  Version:     0.5 28/12/95
 *  Description: Provides a GUI to introduce mathematical
 *               symbols in lyx.
 *
 *  Dependencies: Xlib, XForms, Lyx
 *
 *  Copyright: (c) 1995, Alejandro Aguilar Sierra 
 *
 *   This is the latest version for Lyx public release 0.8.
 *   Next version will be within the math editor for Lyx.
 * 
 *   You are free to use and modify it under the terms of
 *   the GNU General Public Licence version 2 or later.
 */

#include "bmtable.h"                 

#ifndef MATHSYMDEMO
#include "lyx.h"
#include "file.h"
#include "lyx.h"
#include "spellchecker.h"
#include "lyxserver.h"
#include <ctype.h>
#include <unistd.h>
#include <signal.h>   

extern Buffer *buffer[];
extern int current_buffer;
extern void SmallUpdate(signed char);
extern void BeforeChange();
extern void Update(signed char);
#endif

enum  {
   MM_GREEK, MM_ARROW, MM_BOP, MM_BRELATS, MM_VARSIZE, MM_MISC, MM_MAX
};

typedef struct {
	FL_FORM *form;
	FL_OBJECT *greek;
	FL_OBJECT *arrow;
	FL_OBJECT *boperator;
	FL_OBJECT *brelats;
	FL_OBJECT *varsize;
	FL_OBJECT *misc;
	void *vdata;
	long ldata;
} FD_form;

/* Bitmaps */
#include "greek.xbm"
#include "arrows.xbm"
#include "brel.xbm"
#include "bop.xbm"
#include "misc.xbm"           
#include "varsz.xbm"

/* Latex code for those bitmaps */
static char *latex_greek[] =  {
   "Gamma", "Delta", "Theta", "Lambda", "Xi", "Pi",
   "Sigma", "Upsilon", "Phi", "Psi", "Omega",
   "alpha", "beta", "gamma", "delta", "epsilon", "varepsilon", "zeta",
   "eta", "theta", "vartheta", "iota", "kappa", "lambda", "mu",
   "nu", "xi", "pi", "varpi", "rho", "sigma", "varsigma",
   "tau", "upsilon", "phi", "varphi", "chi", "psi", "omega"
};

static char *latex_brel[] = {
  "leq", "geq", "equiv", "models", 
  "prec", "succ", "sim", "perp", 
  "preceq", "succeq", "simeq", "mid", 
  "ll", "gg", "asymp", "parallel", 
  "subset", "supset", "approx", "smile", 
  "subseteq", "supseteq", "cong", "frown", 
  "sqsubseteq", "sqsupseteq", "doteq", "neq", 
  "in", "ni", "propto", "notin", 
  "vdash", "dashv", "bowtie"
};

static char* latex_arrow[] = {
  "downarrow", "leftarrow", "Downarrow", "Leftarrow", 
  "hookleftarrow", "rightarrow", "uparrow", "Rightarrow", "Uparrow",
  "hookrightarrow","updownarrow", "Leftrightarrow", "leftharpoonup", 
  "rightharpoonup", "rightleftharpoons", "leftrightarrow", "Updownarrow", 
  "leftharpoondown", "rightharpoondown", "mapsto",
  "Longleftarrow", "Longrightarrow", "Longleftrightarrow", 
  "longleftrightarrow", "longleftarrow", "longrightarrow", "longmapsto",
  "nwarrow", "nearrow", "swarrow", "searrow",  
};

char *latex_varsz[] = {
"sum", "int", "oint", 
"prod", "coprod", "bigsqcup", 
"bigotimes", "bigodot", "bigoplus", 
"bigcap", "bigcup", "biguplus", 
"bigvee", "bigwedge"
};

static char* latex_bop[] = {   
  "pm", "cap", "diamond", "oplus", 
  "mp", "cup", "bigtriangleup", "ominus", 
  "times", "uplus", "bigtriangledown", "otimes", 
  "div", "sqcap", "triangleright", "oslash", 
  "cdot", "sqcup", "triangleleft", "odot", 
  "star", "vee", "amalg", "bigcirc", 
  "setminus", "wedge", "dagger", "circ", 
  "bullet", "wr", "ddagger"
};

static char* latex_misc[] = {
  "nabla", "partial", "infty", "prime", "ell", 
  "emptyset", "exists", "forall", "i",  "j",
  "Re", "Im", "aleph", "wp", "hbar", 
  "angle", "top", "bot", "Vert", "neg", 
  "flat", "natural", "sharp", "surd", "triangle", 
  "diamondsuit", "heartsuit", "clubsuit", "spadesuit" 
};

typedef FL_OBJECT* FL_OBJECTP;

static void math_cb(FL_OBJECT*, long);
static FD_form* symb_form=0;

// Class to manage bitmap menu bars
class BitmapMenu {
   static BitmapMenu *active;
   friend int peek_event(FL_FORM *, void *);
 protected:
   BitmapMenu *next, *prev;
   int nb;             // Number of bitmaps
   int i;              // Current bitmap
   int ww;             // Border width
   int x, y, w, h;
   FL_FORM *form;
   FL_OBJECTP *bitmap;
   FL_OBJECT *button;
 public:
   BitmapMenu(int n, FL_OBJECT* bt, BitmapMenu* prevx=NULL);
   ~BitmapMenu();
   FL_OBJECT* AddBitmap(int id, int nx, int ny, int bw, int bh, char* data, Bool vert=True);
   void Create(void);   
   void Hide(void);
   void Show(void);
   void Prev(void);
   void Next(void);
   int  GetIndex(FL_OBJECT* ob);
};

BitmapMenu *BitmapMenu::active = NULL;

BitmapMenu::BitmapMenu(int n,  FL_OBJECT* bt, BitmapMenu* prevx): nb(n)  {
   w = h = 0;
   form = 0;
   i = 0;
   ww = 2*FL_abs(FL_BOUND_WIDTH);
   x = y = ww;
   y += 8;
   bitmap = new FL_OBJECTP[nb];
   button = bt;
   button->u_vdata = this;
   prev = prevx;
   next = NULL;
   if (prev)
     prev->next = this;
}

inline
BitmapMenu::~BitmapMenu() {
 if (next) delete next;
 fl_free_form(form);  
 delete[] bitmap;
}

inline
void BitmapMenu::Hide()  {
   fl_hide_form(form);
   fl_set_button(button, 0);
   active = NULL;
}

inline
void BitmapMenu::Show()  {
   if (active)
     active->Hide();
   active = this;
   int x = button->form->x + button->x, y = button->form->y + button->y;
   fl_set_form_position(form, x, y + button->h);
   fl_set_button(button, 1);
   fl_show_form(form, FL_PLACE_POSITION, FL_NOBORDER, "");
}

inline
void BitmapMenu::Prev()  {
   Hide();
   if (prev)
     prev->Show();
}

inline
void BitmapMenu::Next()  {
   Hide();
   if (next)
     next->Show();
}

FL_OBJECT*
BitmapMenu::AddBitmap(int id, int nx, int ny, int bw, int bh, char* data, Bool vert)
{
   if (i>=nb)
     return NULL;
   int wx=bw+ww/2, wy=bh+ww/2;
   wx += (wx % nx);
   wy += (wy % ny); 
   FL_OBJECT *obj = fl_create_bitmaptable(1, x, y, wx, wy, "");   
   fl_set_object_callback(obj, math_cb, id);
   fl_set_object_lcol(obj, FL_BLUE);    
   fl_set_object_boxtype(obj, FL_UP_BOX);   
   fl_set_bitmaptable_data(obj, nx, ny, bw, bh, data);
   if (vert) { 
      y += wy + 8;
      h = FL_max(y, h);
      w = FL_max(x + wx + ww, w);
   } else  {
      x += wx + 8;
      w = FL_max(x, w);
      h = FL_max(y + wy + ww, h);
   }
   bitmap[i++] = obj;
   return obj;
}

void BitmapMenu::Create()
{
   if (i<nb)  {
      fprintf(stderr, "Error: Bitmaps not created!");
      return;
   }
   form = fl_bgn_form(FL_UP_BOX, w, h);   
   for (i=0; i<nb; i++) {
      fl_add_object(form, bitmap[i]);
      bitmap[i]->u_vdata = this;
   }
   fl_end_form();
}

int BitmapMenu::GetIndex(FL_OBJECT* ob)
{
   if (active==this) {
      int k = 0;
      for (i=0; i<nb; i++) {
	 if (bitmap[i]==ob) 
	   return k+fl_get_bitmaptable(ob);
	 k += fl_get_bitmaptable_maxitems(bitmap[i]);
      }
   }
   return -1;
}

int peek_event(FL_FORM *form, void *xev)
{
   if (BitmapMenu::active==NULL)
     return 0;
   
   if(((XEvent *)xev)->type == ButtonPress)
   {
	 BitmapMenu::active->Hide();
	 return 1;
   }
   if(((XEvent *)xev)->type == KeyPress)
   {
      char c[5];
      KeySym keysym;
      XLookupString(&((XEvent *)xev)->xkey, &c[0], 5, &keysym, NULL);
      if (keysym==XK_Left) 
	BitmapMenu::active->Prev(); else
      if (keysym==XK_Right) 
	BitmapMenu::active->Next(); 
      else 
	BitmapMenu::active->Hide();
      return 1;
   }
   return 0;  
}

static void math_cb(FL_OBJECT* ob, long data)
{
   BitmapMenu* menu = ob->u_vdata;
   int i = menu->GetIndex(ob);   
   char *s = 0;
   
   if (i<0) return;
   switch (data)  {
    case MM_GREEK: 
      s = latex_greek[i]; 
      break;
    case MM_ARROW: 
      s = latex_arrow[i]; 
      break;  
    case MM_BRELATS: 
      s = latex_brel[i]; 
      break;  
    case MM_BOP: 
      s = latex_bop[i]; 
      break;  
    case MM_VARSIZE: 
      s = latex_varsz[i];  
      break;
    case MM_MISC: 
      s = latex_misc[i];  
      break;
   }
  
   if (s)  {
#ifdef MATHSYMDEMO      
      fprintf(stderr, "%s", s);  fflush(stderr);
#else
      if (buffer[current_buffer]->text)   {
	 char sx[80];
	 sprintf(sx,"\\%s ",s);
	 buffer[current_buffer]->screen->HideCursor();
	 BeforeChange();
	 Update(-2);
	 buffer[current_buffer]->text->current_font.latex = LYX_FORMULA_LATEX;	     
	 InsetFormulaLatex *new_inset = new InsetFormulaLatex(sx);
	 buffer[current_buffer]->text->InsertInset(new_inset);
	 buffer[current_buffer]->text->sel_cursor = buffer[current_buffer]->text->cursor;
	 SmallUpdate(-1);
	 Update(0);
      }
#endif
   }      
   if (menu)  
     menu->Hide(); 
}

static BitmapMenu* sym_menu=0;

void  create_symbol_menues(void)
{
   FL_OBJECT* obj; 
   BitmapMenu* menu;
   
   sym_menu = menu = new BitmapMenu(2, symb_form->greek);
   obj = menu->AddBitmap(MM_GREEK, 6, 2, Greek_width, Greek_height, 
			   Greek_bits);         
   fl_set_bitmaptable_maxitems(obj, 11);
   obj = menu->AddBitmap(MM_GREEK, 7, 4, greek_width, greek_height,
			   greek_bits); 
   menu->Create();
   
   menu = new BitmapMenu(1, symb_form->boperator, menu);
   obj = menu->AddBitmap(MM_BOP, 4, 8, bop_width, bop_height,
                         bop_bits);      
   fl_set_bitmaptable_maxitems(obj, 31);
   menu->Create();   
      
   menu = new BitmapMenu(1, symb_form->brelats, menu);
   obj = menu->AddBitmap(MM_BRELATS, 4, 9, brel_width, brel_height, 
			 brel_bits);        
   fl_set_bitmaptable_maxitems(obj, 35);       
   menu->Create();
   
   menu = new BitmapMenu(3, symb_form->arrow, menu);
   obj = menu->AddBitmap(MM_ARROW, 5, 4, arrow_width, arrow_height,
			 arrow_bits);
   obj = menu->AddBitmap(MM_ARROW, 2, 4, larrow_width, larrow_height, 
			 larrow_bits, False);          
   fl_set_bitmaptable_maxitems(obj, 7);
   obj = menu->AddBitmap(MM_ARROW, 2, 2, darrow_width, darrow_height, 
			 darrow_bits);
   menu->Create(); 
     
   menu = new BitmapMenu(1, symb_form->varsize, menu);
    obj = menu->AddBitmap(MM_VARSIZE, 3, 5, varsz_width, varsz_height, 
			 varsz_bits);         
   fl_set_bitmaptable_maxitems(obj, 14);
   menu->Create();
      
   menu = new BitmapMenu(1, symb_form->misc, menu);
   obj = menu->AddBitmap(MM_MISC, 5, 6, misc_width, misc_height,
                         misc_bits);      
   fl_set_bitmaptable_maxitems(obj, 29);
   menu->Create();
}

void free_symbols_form()
{
   if (symb_form) {
      fl_hide_form(symb_form->form);
      fl_free_form(symb_form->form);
      delete sym_menu;
      delete symb_form;
      symb_form = NULL;  
   }
}

int AtClose_symbols_form(FL_FORM* f, void *d)
{
  free_symbols_form();
  return FL_IGNORE;
}

static void button_cb(FL_OBJECT *ob, long data)
{
   BitmapMenu *menu = 0;
   switch (data)  {
    case MM_GREEK:
    case MM_VARSIZE:
    case MM_BRELATS:
    case MM_ARROW:
    case MM_BOP:
    case MM_MISC: 
      menu = (BitmapMenu *)ob->u_vdata;
      menu->Show();  
      break;
    case 100:
      free_symbols_form();
      break;
   }
}

FD_form *create_symb_form(void)
{
  FL_OBJECT *obj;
  FD_form *fdui = (FD_form *) fl_calloc(1, sizeof(FD_form));
  int old_bw = fl_get_border_width();

  fl_set_border_width(-2);
  fdui->form = fl_bgn_form(FL_NO_BOX, 332, 90);
  obj = fl_add_box(FL_UP_BOX,0,0,332,90,"");
  fdui->greek = obj = fl_add_button(FL_NORMAL_BUTTON,11,12,50,30,"Greek");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,15);
    fl_set_object_callback(obj,button_cb,MM_GREEK);
  fdui->arrow = obj = fl_add_button(FL_NORMAL_BUTTON,162,12,50,30," ");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,15);
    fl_set_object_callback(obj,button_cb,MM_ARROW);
  fdui->boperator = obj = fl_add_button(FL_NORMAL_BUTTON,62,12,50,30," ");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,15);
    fl_set_object_callback(obj,button_cb,MM_BOP);
  fdui->brelats = obj = fl_add_button(FL_NORMAL_BUTTON,112,12,50,30," @");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,15);
    fl_set_object_callback(obj,button_cb,MM_BRELATS);
  fdui->varsize = obj = fl_add_button(FL_NORMAL_BUTTON,212,12,50,30,"S  ");
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,15);
    fl_set_object_callback(obj,button_cb,MM_VARSIZE);
  fdui->misc = obj = fl_add_button(FL_NORMAL_BUTTON,262,12,50,30,"Misc");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_TIMESITALIC_STYLE);
    fl_set_object_callback(obj,button_cb,MM_MISC);
  obj = fl_add_button(FL_RETURN_BUTTON,246,60,63,21,"Hide ");
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
#ifndef MATHSYMDEMO
   fl_set_object_callback(obj, button_cb, 100);
#endif   
  obj = fl_add_text(FL_NORMAL_TEXT,6,62,110,20,"Math Symbols");
    fl_set_object_lcol(obj,FL_BOTTOM_BCOL);
    fl_set_object_lsize(obj,FL_LARGE_SIZE);
    fl_set_object_lstyle(obj,FL_ITALIC_STYLE+FL_ENGRAVED_STYLE);
  fl_end_form();
  fl_set_border_width(old_bw);

  return fdui;
}

void show_symbols_form(void)
{
   if (!symb_form) {
      symb_form = create_symb_form();
      fl_register_raw_callback(symb_form->form, 
			       ButtonPressMask|KeyPressMask, peek_event);
      create_symbol_menues();	
#ifndef MATHSYMDEMO
      fl_set_form_atclose(symb_form->form, AtClose_symbols_form, NULL);
#endif 
   }
   fl_show_form(symb_form->form, FL_PLACE_SIZE, FL_FULLBORDER, "Math Panel");
}

#ifdef MATHSYMDEMO
int main(int argc, char *argv[])
{
   fl_initialize(argv[0], "Bitmap Table Demo", 0, 0, &argc, argv);	    
   show_symbols_form();                 
   fl_do_forms();
   free_symbols_form();
}     
#endif

