/* This file is part of
* ======================================================
* 
*           LyX, the High Level Word Processor
* 	 
*	    Copyright (C) 1995 Matthias Ettrich
*
*======================================================*/

#include "inset.h"
#include "layout.h"
#include <ctype.h>


#define LYX_PAPER_MARGIN 20

extern Widget toplevel;
extern  WriteAlert(const char* s1, const char* s2, const char* s3);
extern  SetMinibuffer(const char* s1, const char* s2, const char* s3);

static int IsNewlineChar(char c) {
   return (c == LYX_META_NEWLINE);
}

static int IsHfillChar(char c) {
   return (c == LYX_META_HFILL);
}


static int IsSeparatorChar(char c) {
   return (c == ' ' || c == LYX_META_PROTECTED_SEPARATOR);
}

static int IsLineSeparatorChar(char c) {
   return (c == ' ');
}

static int IsKommaChar(char c) {
   return (c == ',' 
	   || c=='('
	   || c==')'
	   || c=='['
	   || c==']'
	   || c==LYX_META_INSET
	   || c==';'
	   || c=='.'
	   || c==':'
	   || c=='-'
	   || c=='?'
	   || c=='!'
	   || c=='&'
	   || c=='@'
	   || c=='+'
	   || c=='-'
	   || c=='~'
	   || c=='#'
	   || c=='%'
	   || c=='^'
	   || c=='/' 
	   || c==LYX_META_NEWLINE
	   || c==LYX_META_PROTECTED_SEPARATOR
	   );
}

static int IsNewline(LyXParagraph  *par, int pos) {
  if (pos<0)
    return 0;
  
  return IsNewlineChar(par->GetChar(pos));
}

static int IsHfill(LyXParagraph  *par, int pos) {
   return IsHfillChar(par->GetChar(pos));
}

static int IsSeparator(LyXParagraph  *par, int pos) {
   return IsSeparatorChar(par->GetChar(pos));
}

static int IsLineSeparator(LyXParagraph  *par, int pos) {
   return IsLineSeparatorChar(par->GetChar(pos));
}
static int IsKomma(LyXParagraph  *par, int pos) {
   return IsKommaChar(par->GetChar(pos));
}



static int FontEqual(LyXFont font1, LyXFont font2) {
   return (font1.family == font2.family 
	   && font1.series == font2.series
	   && font1.shape == font2.shape
	   && font1.size == font2.size
	   && font1.latex == font2.latex
	   && font1.bar == font2.bar);
}







int LyXText::SingleWidth(LyXParagraph *par, int pos) {
    int fsn;
   /* need inset support later */
   
   char c;
   LyXFont font;
   
   c = par->GetChar(pos);
   
   if (IsNewlineChar(c))
     return 0; 
   
   if (IsHfillChar(c))
     return 3;			       /* because of the representation
					* as vertical lines */
   

   font = GetFont(par, pos);
   
   if (c == LYX_META_FOOTNOTE ||
       c == LYX_META_MARGIN ||
       c == LYX_META_FIG ||
       c == LYX_META_TAB) {
	  char* fs;
	  switch (c) {
	   case LYX_META_MARGIN:
	     fs = "margin\0";
	     break;
	   case LYX_META_FIG:
	     fs = "fig\0";
	     break;
	   case LYX_META_TAB:
	     fs = "tab\0";
	     break;
	   default:		       /* LYX_META_FOOTNOTE */
	     fs = "foot\0";
	     break;
	  }
	  for (fsn=0; fs[fsn]; fsn++); /* get length of string  */
	  if (font.size > LYX_SIZE_TINY)
	    font.size--;
	  if (font.size > LYX_SIZE_TINY)
	    font.size--;
	  return LyXTextWidth(font, fs, fsn);
       }
   
   
   if (c == LYX_META_INSET) {
      Inset *tmpinset=par->GetInset(pos);
      if (tmpinset)
      	return par->GetInset(pos)->Width(font);
      else
      	return 0;
   }
   
   
   
   if (IsSeparatorChar(c))
     c = ' ';
   
   if (font.shape == LYX_SMALLCAPS_SHAPE) {
      /* there does not exist a X11 Font with smallcaps shape. So I 
       * have to emulate it */
      if (islower(c)) {
	 /* it is a lowercase letter. so take a smaller font when possible 
	  * and draw it a capital letter */
	 if (font.size > LYX_SIZE_TINY)
	   font.size--;
	 c = toupper(c);
      }
      /* draw it with upshape */
      font.shape = LYX_UP_SHAPE;
   }
   return LyXWidth(font, c);
}





static void DrawLine(Pixmap pm, int baseline, int x, int length) {
   XDrawLine(fl_display,
	     pm,
	     LyXGetGreenGC(),
	     x, baseline,
	     x + length, baseline);
   
   
}

static void DrawOnOffLine(Pixmap pm, int baseline, int x, int length) {
   XDrawLine(fl_display,
	     pm,
	     LyXGetOnOffLineGC(),
	     x, baseline,
	     x + length, baseline);
}

static void DrawThickLine(Pixmap pm, int baseline, int x, int length) {
   XDrawLine(fl_display,
	     pm,
	     LyXGetThickLineGC(),
	     x, baseline,
	     x + length, baseline);
}



static void DrawVerticalLine(Pixmap pm, int x, int y1, int y2) {
   XDrawLine(fl_display,
	     pm,
	     LyXGetGreenGC(),
	     x, y1,
	     x, y2);
   
}

static void DrawVerticalOnOffLine(Pixmap pm, int x, int y1, int y2) {
   XDrawLine(fl_display,
	     pm,
	     LyXGetOnOffLineGC(),
	     x, y1,
	     x, y2);
   
}

/* returns the paragraph position of the last character in the 
* specified row */
int LyXText::RowLast(Row *row) {
   if (row->next == NULL)
     return row->par->Last()-1;
   else if (row->next->par != row->par) 
     return row->par->Last()-1;
   else 
     return row->next->pos - 1;
}


void LyXText::Draw(Row *row, int &pos, unsigned long  pm, float &x) {
   char c;
   LyXFont font;
   int tmppos;
   int last;
   int fsn;
   float tmpx = x;
   
   
   c = row->par->GetChar(pos);
   
   if (IsNewlineChar(c)) {
      pos++;
     return;
   } 
   
   tmppos = pos;
   last = RowLast(row);
   font = GetFont(row->par, pos);
   LyXFont font2 = font;

   
   if (c == LYX_META_FOOTNOTE ||
       c == LYX_META_MARGIN ||
       c == LYX_META_FIG ||
       c == LYX_META_TAB) {
	  char* fs;
	  switch (c) {
	   case LYX_META_MARGIN:
	     fs = "margin\0";
	     /* draw a sign at the left border! */ 
	     XDrawString(fl_display,
			 pm,
			 LyXGetGC(font),
			 (LYX_PAPER_MARGIN - LyXWidth(font, '!'))/2,
			 row->baseline,
			 "!", 1);
	     break;
	   case LYX_META_FIG:
	     fs = "fig\0";
	     break;
	   case LYX_META_TAB:
	     fs = "tab\0";
	     break;
	   default:		       /*  LYX_META_FOOTNOTE:  */
	     fs = "foot\0";
	     break;
	     
	  }
	  for (fsn=0; fs[fsn]; fsn++); /* get length of string  */
	  if (font.size > LYX_SIZE_TINY)
	    font.size--;
	  if (font.size > LYX_SIZE_TINY)
	    font.size--;
	  
	  /* calculate the position of the footnotemark */
	  int y = (row->baseline - LyXMaxAscent(font2) 
		   + LyXAscent(font, 'f'));
	  

	  /* a footnote environement cannot be plain LaTeX, so just
	   * change the font */
	  
	  font.latex = LYX_NO_LATEX;
	  
	  /* draw it! */ 
	  XDrawString(fl_display,
		      pm,
		      LyXGetGC(font),
		      (int) x, y,
		      fs, fsn);
	  
	  /* calculate the width and set the new x position */ 
	  x += LyXTextWidth(font, fs, fsn);
	  
	  pos++;
	  return;
       }  
   
   if (c == LYX_META_INSET) {
      Inset *tmpinset = row->par->GetInset(pos);
      if (tmpinset) 
	tmpinset->Draw(font, pm, row->baseline, x);
      pos++;
      return;
   }
   else {
     /* usual characters, no insets */
     
     if (font.shape == LYX_SMALLCAPS_SHAPE) {
       /* there does not exist a X11 Font with smallcaps shape. So I 
	* have to emulate it */ 
       if (islower(c)) {
	 /* it is a lowercase letter. so take a smaller font when possible 
	  * and draw it a capital letter */
	 if (font.size > LYX_SIZE_TINY)
	   font.size--;
	 c = toupper(c);
       }
       /* draw it with upshape */
       font.shape = LYX_UP_SHAPE;
     }
     else {
       /* select character that you can draw in one command (faster!!)
	* This is not done with smallcaps, of cause! */
       while (pos < last  &&
	      (isalpha(row->par->GetChar(pos + 1)) &&
	       isalpha(row->par->GetChar(pos)) &&
	       FontEqual(font2, GetFont(row->par, pos + 1))))
	 pos++;
     }
     pos++; 			       /* the next character,
					* NOT printed this time */
   
   /* copy the selected characters into a vertex */ 
     char * textstring = new char[pos - tmppos];
     int i;
     textstring[0] = c;
     for (i=1; i < pos - tmppos; i++)
       textstring[i] = row->par->GetChar(tmppos + i);
   
   /* draw it! */ 
     XDrawString(fl_display,
		 pm,
		 LyXGetGC(font),
		 (int) x, row->baseline,
		 textstring, pos - tmppos);

     /* calculate the width and set the new x position */ 
     x +=  LyXTextWidth(font, textstring, pos - tmppos);
   
   
     /* free it */ 
     delete(textstring);
   }
   
   /* what about underbars? */
   if (font.bar == LYX_UNDER_BAR
       && font.latex == LYX_NO_LATEX){
     XDrawLine(fl_display,
	       pm,
	       LyXGetCopyGC(),
	       (int)tmpx, row->baseline + 2,
	       (int)x, row->baseline + 2);
   }
   
}




/* Returns the left beginning of the text. 
* This information cannot be taken from the layouts-objekt, because in 
* LaTeX the beginning of the text fits in some cases (for example sections)
* exactly the label-width. */
int LyXText::LeftBorder(Row* row) {
   int x;
   char*  parindent;
   LyXLayout *layout;
   LyXFont labelfont;
   
   LyXParagraph *newpar;
   Row dummyrow;
   layout = style(parameters->textclass, row->par->GetLayout());
   
   parindent = layout->parindent; 

   /* this is the way, LyX handles the LaTeX-Environements.
    * I have had this idea very late, so it seems to be a
    * later added hack and this is true */ 
   if (!row->par->GetDepth()) {
      x = LYX_PAPER_MARGIN;
      if (!row->par->GetLayout()) {
	 /* find the previous same level paragraph */
	 if (row->par->FirstPhysicalPar()->Previous()) {
	    newpar = row->par->DepthHook(row->par->GetDepth());
	    if (newpar && style(parameters->textclass, newpar->GetLayout())->nextnoindent)
	       parindent = "";
	 }
      }
   }
   else {
      /* find the next level paragraph */ 
      
      newpar = row->par->DepthHook(row->par->GetDepth()-1);

      /* make a corresponding row. Needed to call LeftBorder() */
      
      /* check wether it is a sufficent paragraph */ 
      if (newpar && newpar->footnoteflag == row->par->footnoteflag
	  && style(parameters->textclass, newpar->GetLayout())->latextype != LATEX_PARAGRAPH
	  && style(parameters->textclass, newpar->GetLayout())->latextype != LATEX_COMMAND) {
	     dummyrow.par = newpar;
	     dummyrow.pos = newpar->Last();
	     x = LeftBorder(&dummyrow);
	  }
      else {
	 x = LYX_PAPER_MARGIN;
	 /* this is no longer an error, because this function is used
	  * to clear impossible depths after changing a layout. Since there
	  * is always a redo, LeftBorder() is always called */ 
	 
	 /* printf("LYX_ERROR (text, LeftBorder()) impossible depth \n");*/ 
	 row->par->FirstPhysicalPar()->depth = 0;
      }
      
      if (newpar && !row->par->GetLayout()) {
	 if (newpar->FirstPhysicalPar()->noindent)
	    parindent = "";
	 else
	    parindent = style(parameters->textclass, newpar->GetLayout())->parindent;
      }
      
   }
   
   
   labelfont = layout->labelfont;
   switch (layout->bordertype) {
    case BORDER_DYNAMIC:
      if (layout->leftborder[0]) {
	 x += LyXStringWidth(style(parameters->textclass, 0)->font, 
			     layout->leftborder);
      }
      if (row->par->GetLabelString()) {
	    x += LyXStringWidth(labelfont, layout->labelindent);
	    x += LyXStringWidth(labelfont, row->par->GetLabelString());
	    x += LyXStringWidth(labelfont, layout->labelsep);
	 }
      break;
    case BORDER_MANUAL:
      x += LyXStringWidth(labelfont, layout->labelindent);
      if (row->pos >= BeginningOfMainBody(row->par)) {
	 if (row->par->GetLabelWidthString()) {
	    x += LyXStringWidth(labelfont, row->par->GetLabelWidthString());
	    x += LyXStringWidth(labelfont, layout->labelsep);
	 }
      }
      break;
    case BORDER_STATIC:
      x += ( LyXStringWidth(layout->font, layout->leftborder) * 4
	    / (row->par->GetDepth() + 4));
      break;
    case BORDER_FIRST_DYNAMIC:
      if (layout->labeltype == LABEL_MANUAL) {
	 if (row->pos >= BeginningOfMainBody(row->par)) {
	    x += LyXStringWidth(labelfont, layout->leftborder);
	 }
	 else {
	    x += LyXStringWidth(labelfont, layout->labelindent);
	 }
      }
      else {
	 if (row->pos) {
	    x += LyXStringWidth(labelfont, layout->leftborder);
	 }
	 else if (layout->labeltype != LABEL_TOP_ENVIRONMENT){
	    x += LyXStringWidth(labelfont, layout->labelindent);
	    x += LyXStringWidth(labelfont, layout->labelsep);
	    x += LyXStringWidth(labelfont, row->par->GetLabelString());
	 }
      }
      break;
      
    case BORDER_RIGHT_ADDRESS_BOX:
      /* ok, a terrible hack. The left border depends on the widest row
       * in this paragraph. Do not care about footnotes, they are *NOT*
       * allowed in the LaTeX realisation of this layout. */ 
      
      /* find the first row of this paragraph */ 
      Row *tmprow = row;
      while (tmprow->previous && tmprow->previous->par == row->par)
      	tmprow = tmprow->previous;
      
      int minfill = tmprow->fill;
      while (tmprow-> next && tmprow->next->par == row->par) {
	 tmprow = tmprow->next;
	 if (tmprow->fill < minfill)
	   minfill = tmprow->fill;
      }
      
      x += LyXStringWidth(layout->font, layout->leftborder);
      x += minfill;
      break;
   }
   

   int align;
   if (row->par->FirstPhysicalPar()->align == LYX_ALIGN_LAYOUT)
     align = layout->align;
   else
     align = row->par->FirstPhysicalPar()->align;
   
   /* set the correct parindent */
   if (row->pos == 0
       && layout->labeltype == LABEL_NO_LABEL
       && row->par == row->par->FirstPhysicalPar()
       && align == LYX_ALIGN_BLOCK
       && !row->par->noindent
       && (row->par->layout || parameters->paragraph_separation == LYX_PARSEP_INDENT))
      x += LyXStringWidth(layout->font,parindent);
   
   return x;
}
    
   
int LyXText::RightBorder(Row *row) {
   int  x;
   LyXLayout* layout;
   
   LyXParagraph *newpar;
   Row dummyrow;
   layout = style(parameters->textclass, row->par->GetLayout());

   
    /* this is the way, LyX handles the LaTeX-Environements.
    * I have had this idea very late, so it seems to be a
    * later added hack and this is true */ 
   if (!row->par->GetDepth()) {
      x = LYX_PAPER_MARGIN;
   }
   else {
      /* find the next level paragraph */ 
      
      newpar = row->par;
      
      do {
	 newpar = newpar->FirstPhysicalPar()->Previous()->FirstPhysicalPar();
      } while (newpar && newpar->GetDepth() >= row->par->GetDepth()
	       && newpar->footnoteflag == row->par->footnoteflag);
      

      /* make a corresponding row. Needed to call LeftBorder() */

      /* check wether it is a sufficent paragraph */ 
      if (newpar && newpar->footnoteflag == row->par->footnoteflag
	  && style(parameters->textclass, newpar->GetLayout())->latextype != LATEX_PARAGRAPH
	  && style(parameters->textclass, newpar->GetLayout())->latextype != LATEX_COMMAND) {
	     dummyrow.par = newpar;
	     dummyrow.pos = 0;
	     x = RightBorder(&dummyrow);
	  }
      else {
	 x = LYX_PAPER_MARGIN;
	 /* this is no longer an error, because this function is used
	  * to clear impossible depths after changing a layout. Since there
	  * is always a redo, LeftBorder() is always called */ 
	 
	 /* printf("LYX_ERROR (text, LeftBorder()) impossible depth \n");*/ 
	 row->par->FirstPhysicalPar()->depth = 0;
      }
      
   }


   x += (LyXStringWidth(layout->font, layout->rightborder) * 4
	 / (row->par->GetDepth() + 4));
   return x;
   
}



int LyXText::LabelEnd (Row *row) {
   if (style(parameters->textclass, row->par->GetLayout())->bordertype == BORDER_MANUAL) {
      Row tmprow;
      tmprow = *row;
      tmprow.pos = row->par->Last();
      return LeftBorder(&tmprow);      /* just the beginning 
					* of the main body */
   }
   else
     return 0;			       /* LabelEnd is only needed, if the  
					* layout fills a flushleft
					* label. */
}





/* get the next breakpoint in a given paragraph */
int LyXText::NextBreakPoint(Row* row, int width) {
   int x = 0;
   int last_separator = -1; /* position of the last possible breakpoint 
			    * -1 isn't a suitable value, but a flag */
   int i = 0;
   int left_border;
   
   LyXParagraph *par = row->par;
   int pos = row->pos;
   
   left_border = LabelEnd(row);
   width -= RightBorder(row);
   int main_body = BeginningOfMainBody(par);
   LyXLayout* layout = style(parameters->textclass, par->GetLayout());
   i = pos;

   if (layout->bordertype == BORDER_RIGHT_ADDRESS_BOX) {
      /* special code for right address boxes, only newlines count */
      while (i < par->Last()) {
	 
	 if (IsNewline(par, i)) {
	   last_separator = i;
	   i = par->Last() - 1;/* this means break  */
	   x = width;
	 }
	 i++;
      }
      
   }
   else {
      /* this is the usual handling */ 
      x = LeftBorder(row);
      while (x < width && i < par->Last()) {
	 
	 if (IsNewline(par, i)) {
	   last_separator = i;
	   x = width;		       /* this means break  */
	 }
	 else  {
	    if (IsLineSeparator(par, i))
	      last_separator = i;
	    x += SingleWidth(par, i);
	 }
	 i++;
	 if (i == main_body) {
	    x += LyXStringWidth(layout->labelfont,
				layout->labelsep);
	    if (IsLineSeparator(par, i - 1))
	      x-= SingleWidth(par, i - 1);
	    if (x < left_border)
	      x = left_border;
	 }
      }
      /* end of paragraph is always a suitable separator */
     if (i == par->Last() && x < width)
       last_separator = i;
      
   }
   
   
   
   /* well, if last_separator is still 0, the line isn't breakable. 
    * don't care and cut simply at the end */
   if (last_separator < 0) {
     last_separator = i;
   }
   
   /* manual labels cannot be broken in LaTeX, do not care  */
   if (main_body && last_separator < main_body)
     last_separator = main_body - 1;

   return last_separator;
}






/* returns the minimum space a row needs on the screen in pixel */
int LyXText::Fill(Row *row, int paperwidth) {
   int w,i, last, fill, left_border;
   /* get the pure distance */
   last = RowLast(row);

   left_border = LabelEnd(row);
   
   /* if the row ends with newline, this newline will not be relevant */ 
   if (IsNewline(row->par, last))
     last--;

   /* if the row ends with a space, this space will not be relevant */ 
   if (IsLineSeparator(row->par, last))
     last--;
   
   /* special handling of the right address boxes */ 
   if (style(parameters->textclass, row->par->GetLayout())->bordertype == BORDER_RIGHT_ADDRESS_BOX) {
      int tmpfill = row->fill;
      row->fill = 0; 		       /* the minfill in BorderLeft()  */
      w = LeftBorder(row);
      row->fill = tmpfill;
      
      /* the old way, impossible when using environements: */ 
      /*  w = LyXStringWidth(style(parameters->textclass, row->par->GetLayout())->font, */ 
      /* 			 style(parameters->textclass, row->par->GetLayout())->leftborder); */
   }
   else
     w = LeftBorder(row);
   
   int main_body = BeginningOfMainBody(row->par);
   LyXLayout *layout = style(parameters->textclass, row->par->GetLayout());
   i = row->pos;
   while (i<= last) {
      w += SingleWidth(row->par, i);
      i++;
      if (i == main_body) {
	 w += LyXStringWidth(layout->labelfont,
			     layout->labelsep);
	 if (IsLineSeparator(row->par, i - 1))
	   w-= SingleWidth(row->par, i - 1);
	 if (w < left_border)
	   w = left_border;
      }
   }
   
   fill = paperwidth - w - RightBorder(row);
   return fill;
}


/* returns the minimum space a manual label needs on the screen in pixel */ 
int LyXText::LabelFill(Row *row) {
   int w,i, last;
   
   int fill=0;
   
   last = BeginningOfMainBody(row->par) - 1;
   /* -1 because a label ends either with a space that is in the label, 
    * or with the beginnig of a footnote that is outside the label. */ 
   
   if (IsLineSeparator(row->par, last))    /* a sepearator at this end 
					* does not count */
     last--;
   
   w = 0;
   i = row->pos;
   while (i<= last) {
      w += SingleWidth(row->par, i);
      i++;
   }
   
   if (row->par->labelwidthstring) {
      fill = LyXStringWidth(style(parameters->textclass, row->par->GetLayout())->labelfont,
			    row->par->labelwidthstring) - w;
			 }
   
   if (fill < 0)
     fill = 0;
   
   return fill;
}


   






/* returns the number of separators in the specified row. The separator 
* on the very last column doesnt count */ 
int LyXText::NumberOfSeparators(Row *row) {
   int n,p,last;
   
   last = RowLast(row);
   n = 0;
   p = row->pos;
   int main_body = BeginningOfMainBody(row->par);
   if (p < main_body)
     p = main_body;
   for (; p < last; p++) {
      if (IsSeparator(row->par, p)) {
	 n++;
      }
   }
   return n;
}
	 
/* returns the number of hfills in the specified row. The LyX-Hfill is
* a LaTeX \hfill so that the hfills at the beginning and at the end were 
* ignored. This is *MUCH* more usefull than not to ignore!  */
int LyXText::NumberOfHfills(Row *row) {
   int n,p,last, first;
   
   last = RowLast(row);
   first = row->pos;
   if (first) {			       /* hfill *DO* count at the beginning 
					* of paragraphs! */
     while(first < last && IsHfill(row->par, first))
       first++;
   }

   n = 0;
   int main_body = BeginningOfMainBody(row->par);
   if (first < main_body)
     first = main_body;
   for (p = first; p < last; p++) {    /* last, because the end is ignored!  */
      if (IsHfill(row->par, p)) {
	 n++;
      }
   }
   return n;
}

/* like NumberOfHfills, but only those in the manual label! */ 
int LyXText::NumberOfLabelHfills(Row *row) {
   int n,p,last, first;
   
   last = RowLast(row);
   first = row->pos;
   if (first) {			       /* hfill *DO* count at the beginning 
					* of paragraphs! */
      while(first < last && IsHfill(row->par, first))
      	first++;
   }

   n = 0;
   int main_body = BeginningOfMainBody(row->par);
   
   if (last > main_body)
     last = main_body;
   
   for (p = first; p < last; p++) {    /* last, because the end is ignored!  */
      if (IsHfill(row->par, p)) {
	 n++;
      }
   }
   return n;
}



/* boolean. returns true, if a expansion is needed. Rules are given by 
* LaTeX */ 
int LyXText::HfillExpansion(Row *row_ptr, int pos) {
   /* by the way, is it a hfill? */ 
   if (!IsHfill(row_ptr->par, pos))
     return 0;
   
   /* at the end of a row it does not count */ 
   if (pos >= RowLast(row_ptr))
     return 0;
   
   
   /* at the beginning of a row it does not count, if it is not 
    * the first row of a paragaph */ 
   if (!row_ptr->pos)
     return 1;
   
   /* in some labels  it does not count */ 
   if ( style(parameters->textclass, row_ptr->par->GetLayout())->bordertype != BORDER_MANUAL &&
       pos < BeginningOfMainBody(row_ptr->par))
     return 0; 
   
   
   /* if there is anything between the first char of the row and
    * the sepcified position that is not a newline and not a hfill,
    * the hfill will count, otherwise not */ 
   
   int i = row_ptr->pos;
   while (i < pos && (IsNewline(row_ptr->par, i)
		      || IsHfill(row_ptr->par,i)))
     i++;
   
   return (i != pos);
}



/* this will incalculate the layout later!!! */ 
void LyXText::SetHeightOfRow(Row *row_ptr) {
    /* get the maximum ascent and the maximum descent */
   int asc, maxasc, desc, maxdesc, pos_end, pos, labeladdon;
   float layoutasc = 0;
   float layoutdesc = 0;
   float tmptop;
   LyXParagraph *par, *firstpar;
   LyXFont font, tmpfont;
   Inset *tmpinset;
   char c;
   char size;

   /* this must not happen before the currentrow for clear reasons.
      so the trick is just to set the current row onto this row */
   long unused_y;
   GetRow(row_ptr->par, row_ptr->pos, unused_y);


   
   /* ok , let us initialize the maxasc and maxdesc value. 
    * This depends in LaTeX of the font of the last character
    * in the paragraph. The hack below is necessary because
    * of the possibility of open footnotes */

   /* corection: only the fontsize does count. The other properties
      are taken from the layoutfont. Nicer on the screen :) */
   
      
   par = row_ptr->par->LastPhysicalPar();
   firstpar = row_ptr->par->FirstPhysicalPar();
   
   LyXLayout *layout = style(parameters->textclass, firstpar->GetLayout());
   
   font = GetFont(par, par->Last()-1);
   size = font.size;
   
   font = layout->font;
   font.size = size;

   maxasc = (int) (LyXMaxAscent(font) * layout->baselinestretch * parameters->baseline);
   maxdesc = (int) (LyXMaxDescent(font) * layout->baselinestretch * parameters->baseline);

   pos_end = RowLast(row_ptr);
   
   asc = 0;
   desc = 0;
   labeladdon = 0;
   
   
   
   for (pos = row_ptr->pos; pos <= pos_end; pos++) {
      /* needs inset support later. this check is, that I do not 
       * calculate unneccesary things. A X Call like Ascent is
       * too expensive! */
      tmpfont = GetFont(row_ptr->par, pos);
      if (tmpfont.size > font.size) {
	 c = row_ptr->par->GetChar(pos);
	 if (c == LYX_META_INSET) {
	    tmpinset = row_ptr->par->GetInset(pos);
	    if (tmpinset) {
	       asc = tmpinset->Ascent(tmpfont);
	       desc = tmpinset->Descent(tmpfont);
	    }
	 }
	 else {
	    asc = LyXAscent(tmpfont, c);
	    desc = LyXDescent(tmpfont, c);
	 }
	 if (asc > maxasc) 
	   maxasc = asc;
	 if (desc > maxdesc)
	   maxdesc = desc;
      }

   }
   
   /* is it a top line? */ 
   if (row_ptr->pos == 0
       && row_ptr->par == firstpar) {
      
      /* some parksips VERY EASY IMPLEMENTATION */ 
      if (parameters->paragraph_separation == LYX_PARSEP_SKIP) {
	 if (layout->latextype == LATEX_PARAGRAPH
	     && firstpar->GetDepth() == 0
	     && firstpar->Previous())
	    maxasc += (int)(layout->parskip * DefaultHeight());
	 else if (firstpar->Previous()
		  && style(parameters->textclass, firstpar->Previous()->GetLayout())->latextype == LATEX_PARAGRAPH
		  && firstpar->Previous()->GetDepth() == 0)
	    maxasc += (int)(style(parameters->textclass, firstpar->Previous()->GetLayout())->parskip * DefaultHeight());
	 
      }
      
      
      /* the paper margins */ 
      if (!row_ptr->par->previous)
      	 maxasc += LYX_PAPER_MARGIN;
      
      /* add the vertical spaces, that the user added */
      if (firstpar->added_space_top)
      	 maxasc += (int) (firstpar->added_space_top * 37);
      
      /* do not forget the DTP-lines! 
       * there height depends on the font of the nearest character */
      if (firstpar->line_top)
      	 maxasc += 2 * LyXAscent(GetFont(firstpar, 0), 'x');
      
      /* and now the pagebreaks */ 
      if (firstpar->pagebreak_top)
      	 maxasc += 3 * DefaultHeight();
      
      
      /* and the vertical fills */ 
      if (firstpar->fill_top)
      	 maxasc += 3 * DefaultHeight();
      
      
      /*  this is special code for the chapter, since the label of this
       * layout is printed in an extra row */ 
      if (layout->labeltype == LABEL_COUNTER_CHAPTER && parameters->secnumdepth>=0) {
	 labeladdon = maxasc + maxdesc + (int) layout->bottomsep * DefaultHeight();
      }
      
      /* sepcial code for the top label */ 
      if (layout->labeltype == LABEL_TOP_ENVIRONMENT && parameters->secnumdepth >=0 &&
	  (!par->Previous() || par->Previous()->GetLayout() != par->GetLayout())) {
	 LyXLayout *top_label_layout = style(parameters->textclass, NumberOfLayout(parameters->textclass, 
										  TextClass(parameters->textclass)->top_label_layout));
	 
	 labeladdon = 
	 (int) (LyXMaxAscent(top_label_layout->labelfont) * top_label_layout->baselinestretch * parameters->baseline)
	 + (int) (LyXMaxDescent(top_label_layout->labelfont) * top_label_layout->baselinestretch * parameters->baseline)
	 + (int) top_label_layout->topsep * DefaultHeight()
	 + (int) ((top_label_layout->bottomsep - layout->topsep) * DefaultHeight());
      }
   
      
      /* and now the layout spaces, for example before and after a section, 
       * or between the items of a itemize or enumerate environement */ 
      
      if (!firstpar->pagebreak_top) {
	 LyXParagraph *prev = row_ptr->par->Previous();
	 if (prev)
	    prev = row_ptr->par->DepthHook(row_ptr->par->GetDepth());
	 if (prev && prev->GetLayout() == firstpar->GetLayout()
	     && prev->GetDepth() == firstpar->GetDepth()
	     && StringEqual(prev->GetLabelWidthString(),
			    firstpar->GetLabelWidthString()))
	   {
	      layoutasc = (layout->itemsep * DefaultHeight());
	   }
	 else if (row_ptr->previous) {
	    tmptop = layout->topsep;
	    
	    if (row_ptr->previous->par->GetDepth() >= row_ptr->par->GetDepth())
	       tmptop-=style(parameters->textclass, row_ptr->previous->par->GetLayout())->bottomsep;
	    
	    if (tmptop > 0)
	       layoutasc = (tmptop * DefaultHeight());
	 }
	     else if (row_ptr-par->line_top){
		tmptop = layout->topsep;
		
		if (tmptop > 0)
		   layoutasc = (tmptop * DefaultHeight());
	     }
	 
	 prev = row_ptr->par->DepthHook(row_ptr->par->GetDepth()-1);
	 if (prev)  {
	    maxasc += (int) (style(parameters->textclass, prev->GetLayout())->parsep * DefaultHeight());
	 }
	 else {
		if (firstpar->Previous()
		    && firstpar->Previous()->GetDepth() == 0
		    && firstpar->Previous()->GetLayout() != firstpar->GetLayout()) {
		   /* avoid parsep */ 
		}
	    else if (firstpar->Previous()){
	       maxasc += (int) (layout->parsep * DefaultHeight());
	    }
	 }
      }
   }
   
   
   
   
   
   
   /* is it a bottom line? */ 
   if (row_ptr->par->ParFromPos(RowLast(row_ptr) + 1) == par
       && (!row_ptr->next || row_ptr->next->par != row_ptr->par)) {     
	  
	  /* the paper margins */ 
	  if (!par->next)
	    maxdesc += LYX_PAPER_MARGIN;
	
	  
	  /* add the vertical spaces, that the user added */
	  if (firstpar->added_space_bottom)
	    maxdesc += (int) (firstpar->added_space_bottom * 37);
	  
	  
	  /* do not forget the DTP-lines! 
	   * there height depends on the font of the nearest character */
	  if (firstpar->line_bottom)
	    maxdesc += 2 * LyXAscent(GetFont(par, par->Last()-1), 'x');
	  
	  
	  /* and now the pagebreaks */
	  if (firstpar->pagebreak_bottom)
	    maxdesc += 3 * DefaultHeight();
	  
	  
	  /* and the vertical fills */
	  if (firstpar->fill_bottom)
	    maxdesc += 3 * DefaultHeight();
	  
	  
	  /* and now the layout spaces, for example before and after a section, 
	   * or between the items of a itemize or enumerate environement */
	  if (!firstpar->pagebreak_bottom && row_ptr->par->Next()) {
	     LyXParagraph *nextpar = row_ptr->par->Next();
	     LyXParagraph *comparepar = row_ptr->par;
	     float usual = 0;
	     float  unusual = 0;
	     
	     if (comparepar->GetDepth() > nextpar->GetDepth()) {
		usual = (style(parameters->textclass, comparepar->GetLayout())->bottomsep * DefaultHeight());
		comparepar = comparepar->DepthHook(nextpar->GetDepth());
		if (comparepar->GetLayout()!=nextpar->GetLayout()
		    ||(!StringEqual(nextpar->GetLabelWidthString(),
				    comparepar->GetLabelWidthString())))
		  unusual = (style(parameters->textclass, comparepar->GetLayout())->bottomsep * DefaultHeight());
		
		if (unusual > usual)
		  layoutdesc = unusual;
		else
		  layoutdesc = usual;
	     }
	     
	     else if (comparepar->GetDepth() ==  nextpar->GetDepth()) {
		
		if (comparepar->GetLayout()!= nextpar->GetLayout()
		    || (!StringEqual(nextpar->GetLabelWidthString(),
				     comparepar->GetLabelWidthString())))
		  layoutdesc = (int)(style(parameters->textclass, comparepar->GetLayout())->bottomsep * DefaultHeight());
	     }
	  }
       }
   
   /* incalculate the layout spaces */ 
   maxasc += (int) (layoutasc * 2 / (2 + firstpar->GetDepth()));
   maxdesc += (int) (layoutdesc * 2 / (2 + firstpar->GetDepth()));
   
   
   
   /* calculate the new height of the text */ 
   height -= row_ptr->height;
   
   row_ptr->height=maxasc+maxdesc+labeladdon;
   row_ptr->baseline=maxasc+labeladdon;
   
   height += row_ptr->height;
   
}






/* appends  the implizit specified paragraph behind the specified row,
* start at the implizit given position */
void LyXText::AppendParagraph(Row *row) {
   int z;
   Row *tmprow;
   char not_ready;
   
   not_ready = 1;
   
   /* get the width of the row */
   z = row->pos;

   
   do {
      /* get the next breakpoint */
      z = NextBreakPoint(row, paperwidth);
      
      
      tmprow = row;
      /* inset the new row */
      if (z < row->par->Last()) {
	 z++;
	 InsertRow(row, row->par, z);
	 row = row->next;
	 row->height = 0;
      }
      else
      	not_ready = 0;
      
      /* set the dimensions of the row */ 
      tmprow->fill = Fill(tmprow, paperwidth);
      SetHeightOfRow(tmprow);
      
   } while (not_ready);
   

   
}


void LyXText::BreakAgain(Row *row) {
   int z;
   char not_ready;
   z = row->pos;
   Row *tmprow, *tmprow2;
   
   not_ready = 1;
   
   do  {
      /* get the next breakpoint */
      z = NextBreakPoint(row, paperwidth);
      
      tmprow = row;
      
      
      if (z < row->par->Last() ) {
	 if (!row->next || (row->next && row->next->par != row->par)) {
	    /* insert a new row */ 
	    z++;
	    InsertRow(row, row->par, z);
	    row = row->next;
	    row->height = 0;
	 }
	 else  {
	    row=row->next;
	    z++;
	    if (row->pos == z)
	      not_ready = 0;	       /* the rest will not change  */
	    else {
	       row->pos = z;
	    }
	 }
      }
      else {
	 /* if there are some rows too much, delete them */
	 /* only if you broke the whole paragraph! */ 
	 tmprow2 = row;
	 while (tmprow2->next && tmprow2->next->par == row->par) {
	    tmprow2 = tmprow2->next;
	 }
	 while (tmprow2 != row) {
	    tmprow2 = tmprow2->previous;
	    RemoveRow(tmprow2->next);
	 }
	 not_ready = 0;
      }
       
      /* set the dimensions of the row */ 
      tmprow->fill = Fill(tmprow, paperwidth);
      SetHeightOfRow(tmprow);
   } while (not_ready);
   
   
}

/* this is just a little changed version of break again */ 
void LyXText::BreakAgainOneRow(Row *row) {
   int z;
   char not_ready;
   z = row->pos;
   Row *tmprow, *tmprow2;
   
   not_ready = 1;
   
   /* get the next breakpoint */
   z = NextBreakPoint(row, paperwidth);
   
   tmprow = row;
   
   
   if (z < row->par->Last() ) {
      if (!row->next || (row->next && row->next->par != row->par)) {
	 /* insert a new row */ 
	 z++;
	 InsertRow(row, row->par, z);
	 row = row->next;
	 row->height = 0;
      }
      else  {
	 row=row->next;
	 z++;
	 if (row->pos == z)
	   not_ready = 0;	       /* the rest will not change  */
	 else {
	    row->pos = z;
	 }
      }
   }
   else {
      /* if there are some rows too much, delete them */
      /* only if you broke the whole paragraph! */ 
      tmprow2 = row;
      while (tmprow2->next && tmprow2->next->par == row->par) {
	 tmprow2 = tmprow2->next;
      }
      while (tmprow2 != row) {
	 tmprow2 = tmprow2->previous;
	 RemoveRow(tmprow2->next);
      }
      not_ready = 0;
   }
   
   /* set the dimensions of the row */ 
   tmprow->fill = Fill(tmprow, paperwidth);
   SetHeightOfRow(tmprow);
   
}

   
   

void LyXText::BreakParagraph() {
   
   LyXLayout *layout = style(parameters->textclass, cursor.par->GetLayout());
   
   /* this is only allowed, if the current paragraph is not empty or caption*/ 
   if (cursor.par->Last() <= 0 && 
       layout->labeltype!=LABEL_SENSITIVE)
     return;
   /* please break alway behind a space */ 
   if (cursor.pos < cursor.par->Last()
       && IsLineSeparator(cursor.par,cursor.pos))
     cursor.pos++;
   
   
   /* break the paragraph */
   cursor.par->BreakParagraph(cursor.pos, layout->latextype >= LATEX_ENVIRONEMENT);
   /* well this is the caption hack since one caption is really enough */
   if (layout->labeltype == LABEL_SENSITIVE){
     if (!cursor.pos)
       cursor.par->SetLayout(0); /* set the new paragraph to standard-layout */
     else
       cursor.par->Next()->SetLayout(0); /* set the new paragraph to standard-layout */
	 
   }
   
   /* if the cursor is at the beginning of a row without prior newline, 
    *  move one row up! 
    * This touches only the screen-update. Otherwise we would may have
    * an empty row on the screen */
   if (cursor.pos && !IsNewline(cursor.row->par, cursor.row->pos -1) &&
       cursor.row->pos == cursor.pos) {
     CursorLeft();
   } 
   
   status = LYX_NEED_MORE_REFRESH;
   refresh_row = cursor.row;
   refresh_y = cursor.y - cursor.row->baseline;
   
   /* Do not forget the special right address boxes */
   if (layout->bordertype == BORDER_RIGHT_ADDRESS_BOX) {
      while (refresh_row->previous &&
	     refresh_row->previous->par == refresh_row->par) {
		refresh_row = refresh_row->previous;
		refresh_y -= refresh_row->height;
	     }
   }
   RemoveParagraph(cursor.row);
   
   /* set the dimensions of the cursor row */
   cursor.row->fill = Fill(cursor.row, paperwidth);

   SetHeightOfRow(cursor.row);
   
   while (cursor.par->Next()->Last()
	  && IsNewline(cursor.par->Next(),0))
     cursor.par->Next()->Erase(0);
   
   
   InsertParagraph(cursor.par->Next(), cursor.row);

   UpdateCounters(cursor.row->previous);
   
   /* this check is necessary. Otherwise the new empty paragraph will
    * be deleted automatically. And it is more friendly for the user! */ 
   if (cursor.pos)
     SetCursor(cursor.par->Next(), 0);
   else
     SetCursor(cursor.par, 0);
   
   if (cursor.row->next)
     BreakAgain(cursor.row->next);


}






void LyXText::OpenFootnote() {
   LyXParagraph *par, *endpar,*tmppar;
   Row *row;
   
   
   par = cursor.par->ParFromPos(cursor.pos);
   
   /* if there is no footnote in this paragraph, just return. */ 
   if (!par->next
       || par->next->footnoteflag != LYX_CLOSED_FOOTNOTE)
     return;
   
   
   /* ok, move the cursor right before the footnote */ 
   

   /* just a little faster than using CursorRight() */
   for (cursor.pos=0; cursor.par->ParFromPos(cursor.pos)!=par; cursor.pos++);
   /* now the cursor is at the beginning of the physical par */
   SetCursor(cursor.par, cursor.pos + cursor.par->ParFromPos(cursor.pos)->last);
   
   

   /* the cursor must be exactly before the footnote */ 
   par = cursor.par->ParFromPos(cursor.pos);
   
   status = LYX_NEED_MORE_REFRESH;
   refresh_row = cursor.row;
   refresh_y = cursor.y - cursor.row->baseline;
   
   tmppar = cursor.par;
   endpar = cursor.par->Next();
   row = cursor.row;
   
   tmppar->OpenFootnote(cursor.pos);
   RemoveParagraph(row);
   /* set the dimensions of the cursor row */
   row->fill = Fill(row, paperwidth);
   SetHeightOfRow(row);
   tmppar = tmppar->Next();
   
   while (tmppar != endpar) {
      if (tmppar) {
	 InsertParagraph(tmppar, row);
	 while (row->next && row->next->par == tmppar)
	   row = row->next;
	 tmppar = tmppar->Next();

      }
   }
   
   
   
   SetCursor(par->next, 0);
   sel_cursor = cursor;
}
   






/* insert a character, moves all the following breaks in the 
* same Paragraph one to the right and make a rebreak */
void  LyXText::InsertChar(char c) {
   Row *row;
   Row *tmprow;
   int z;
   long y;
   LyXFont tmpfont, realtmpfont, reference_font1, reference_font2;
   char jumped_over_space;	       /* boolean flag  */
   
   
   /* first check, if there will be two blanks together or a blank at 
    * the beginning of a paragraph. 
    * I decided to handle blanks like normal characters, the main 
    * difference are the special checks when calculating the row.fill
    * (blank does not count at the end of a row) and the check here */ 

   realtmpfont = real_current_font;
   tmpfont = current_font;	       /* store the current font.
					* This is because of the use
					* of cursor movements. The moving
					* cursor would refresh the 
					* current font */
   
   /* get the font that is used to calculate the baselineskip */ 
   reference_font1 = GetFont(cursor.par, cursor.par->Last() - 1);
   reference_font2 = cursor.par->GetFontSettings(cursor.par->Last() - 1);
   jumped_over_space = False;
   
   if (IsLineSeparatorChar(c)) {
	   
      
      if (cursor.pos < cursor.par->Last() &&
	       IsLineSeparator(cursor.par,cursor.pos))
 	{
	   /* the user inserted a space before a space. So we
	    * will just make a CursorRight. BUT: The font of this
	    * space should be set to current font. That is why
	    * we need to rebreak perhaps. If there is a protected
	    * blank at the end of a row we have to force a rebreak.*/ 
	   
	   if (cursor.pos == RowLast(cursor.row)
	       && !IsLineSeparatorChar(c))
	     cursor.row->fill = -1;    /* force rebreak  */
	   
	   cursor.par->Erase(cursor.pos);
	   jumped_over_space = True;
	   
	}
      else if ((cursor.pos > 0 && 
		     IsLineSeparator(cursor.par,cursor.pos - 1))
		    || (cursor.pos > 0 && IsNewline(cursor.par, cursor.pos - 1))
		    || (cursor.pos == 0 &&
			!(cursor.par->Previous()
			  && cursor.par->Previous()->footnoteflag
			  == LYX_OPEN_FOOTNOTE)))
      	return;
   }

   else if (IsNewlineChar(c)) {
     if (cursor.par->FirstPhysicalPar() == cursor.par
	 && cursor.pos <= BeginningOfMainBody(cursor.par))
       return;			       /* no newline at first position 
					* of a paragraph or behind labels. 
					* TeX does not allow that. */
      
      if (cursor.pos < cursor.par->Last() &&
	  IsLineSeparator(cursor.par,cursor.pos))
      	CursorRight();		       /* newline always after a blank! */
      cursor.row->fill = -1;	       /* to force a new break  */
   }
   
   
   
   /* get the cursor row fist */
   /* this is a dump solution, i will try to hold the cursor.row in future */ 
   /* row = GetRow(cursor.par, cursor.pos, y);*/
   /* ok, heres a better way: */ 
   row = cursor.row;
   y = cursor.y - row->baseline;
   if (c != LYX_META_INSET)	       /* in this case LyXText::InsertInset 
					* already insertet the character */
     cursor.par->InsertChar(cursor.pos, c);
   SetCharFont(cursor.par, cursor.pos, tmpfont);
   
   if (!jumped_over_space) {
      /* refresh the positions */
      tmprow = row;
      while (tmprow->next && tmprow->next->par == row->par) {
	 tmprow = tmprow->next;
	 tmprow->pos++;
      }
   }
   
   
   
   /* is there a break one row above */ 
   if ((IsLineSeparator(cursor.par, cursor.pos)
	|| IsNewline(cursor.par, cursor.pos))&& 
       row->previous && row->previous->par == row->par) {
	  z = NextBreakPoint(row->previous, paperwidth);
	  if ( z >= row->pos) {
	     row->pos = z + 1;
	     
	     	     
	     /* set the dimensions of the row above  */ 
	     row->previous->fill = Fill(row->previous, paperwidth);

	     SetHeightOfRow(row->previous);
	     
	     y -= row->previous->height;
	     refresh_y = y;
	     refresh_row = row->previous;
	     status = LYX_NEED_MORE_REFRESH;
	     
	     BreakAgainOneRow(row);
	     SetCursor(cursor.par, cursor.pos + 1);
	     /* cursor MUST be in row now */
	     
	     if (row->next && row->next->par == row->par)
	       need_break_row = row->next;
	     else
	       need_break_row = NULL;
	     current_font = tmpfont; real_current_font = realtmpfont;
	     
	     /* check, wether the last characters font has changed. */ 
	     if (cursor.pos && cursor.pos == cursor.par->Last()
		 && !(FontEqual(reference_font1, tmpfont)
		      || FontEqual(reference_font2, tmpfont)))
	       RedoHeightOfParagraph(cursor);
	     
   	     return;
	  }
       }
   
   /* recalculate the fill of the row */ 
   if (row->fill >= 0)		       /* needed because a newline
					* will set fill to -1. Otherwise
					* we would not get a rebreak! */
     row->fill = Fill(row, paperwidth);
   if (row->fill < 0 ) {
      refresh_y = y;
      refresh_row = row; 
      refresh_x = cursor.x;
      refresh_pos = cursor.pos;
      status = LYX_NEED_MORE_REFRESH;
      BreakAgainOneRow(row); 
      /* will the cursor be in another row now? */ 
      if (RowLast(row) <= cursor.pos + 1 && row->next) {
	 if (row->next && row->next->par == row->par)   /* this should
							 * always be true */
	   row = row->next;
	 BreakAgainOneRow(row);
      }
      SetCursor(cursor.par, cursor.pos + 1);
      if (row->next && row->next->par == row->par)
      	need_break_row = row->next;
      else
      	need_break_row = NULL;
      current_font = tmpfont; real_current_font = realtmpfont;
     
   }  
   else  {
      
      refresh_y = y;
      refresh_x = cursor.x;
      refresh_row = row;
      refresh_pos = cursor.pos;
      
      int tmpheight = row->height;
      SetHeightOfRow(row);
      if (tmpheight == row->height)
      	status = LYX_NEED_VERY_LITTLE_REFRESH;
      else
      	status = LYX_NEED_MORE_REFRESH;
      
      
      SetCursor(cursor.par, cursor.pos + 1);
      current_font = tmpfont; real_current_font = realtmpfont;
   }
   
   /* check, wether the last characters font has changed. */ 
   if (cursor.pos && cursor.pos == cursor.par->Last()
       && !(FontEqual(reference_font1, tmpfont)
	    || FontEqual(reference_font2, tmpfont)))
     RedoHeightOfParagraph(cursor);
   else {
      /* now the special right address boxes */
      if (style(parameters->textclass, cursor.par->GetLayout())->bordertype == BORDER_RIGHT_ADDRESS_BOX) {
	 RedoDrawingOfParagraph(cursor); 
      }
   }
   
   
   
}
   
   
   
   
   
   
   
   
   
   







void LyXText::PrepareToPrint(Row *row, float &x, float &fill_separator, 
			     float &fill_hfill, float &fill_label_hfill)
     {
	float w, nh, nlh, ns;
	
	w = row->fill;
	fill_hfill = 0;
	fill_label_hfill = 0;
	fill_separator = 0;
	fill_label_hfill = 0;
	
	x = LeftBorder(row);
	
	/* is there a manual border with a manual label */ 
	if (style(parameters->textclass, row->par->GetLayout())->bordertype == BORDER_MANUAL &&
	    style(parameters->textclass, row->par->GetLayout())->labeltype == LABEL_MANUAL) {
	       
	       nlh = NumberOfLabelHfills(row) + 1; /* one more since labels 
						    * are left aligned*/ 
	       if (nlh && row->par->GetLabelWidthString()) {
		  fill_label_hfill = LabelFill(row) / nlh;
	       }
	    }
	
	
	/* are there any hfills in the row? */ 
	nh = NumberOfHfills(row);
	if (nh)
	  fill_hfill = w /nh;
	else  {
	   /* is it block, flushleft or flushright? 
	    * set x how you need it */
	   int align;
	   if (row->par->FirstPhysicalPar()->align == LYX_ALIGN_LAYOUT)
	     align = style(parameters->textclass, row->par->GetLayout())->align;
	   else
	     align = row->par->FirstPhysicalPar()->align;
	   
	   switch (align) {
	    case LYX_ALIGN_BLOCK:
	      ns = NumberOfSeparators(row);
	      if (ns && row->next && row->next->par == row->par &&
		  !(IsNewline(row->next->par, row->next->pos-1)))
	      	fill_separator = w / ns;
	      break;
	    case LYX_ALIGN_RIGHT:
	      x += w;
	      break;
	    case LYX_ALIGN_FORMULA:
	      if (row->previous && row->previous->par == row->par &&
		  !(IsNewline(row->par, row->pos-1)))
		{
		   /* avoid centering */ 
		}
	      else
	      	x += w / 2;
	      break;
	    case LYX_ALIGN_CENTER:
	      x += w / 2;
	      break;
	   }
	}

     }

      


				



/* important for the screen */


/* the cursor set functions have a special mechanism. When they
* realize, that you left an empty paragraph, they will delete it.
* They also delet the corresponding row */



void  LyXText::SetCursor(LyXParagraph *par, int pos) {
   long y;
   Row *row;
   LyXCursor old_cursor;
   
   old_cursor = cursor;
   int left_border;
   cursor.par = par;
   cursor.pos = pos;
   
   /* get the cursor y position in text  */
   row = GetRow(par, pos, y);
   /* y is now the beginning of the cursor row */ 
   y += row->baseline;
   /* y is now the cursor baseline */ 
   cursor.y = y;
   
   
   /* now get the cursors x position */
   
   float x;
   float fill_separator, fill_hfill, fill_label_hfill;
   left_border = LabelEnd(row);
   PrepareToPrint(row, x, fill_separator, fill_hfill, fill_label_hfill);
   int main_body = BeginningOfMainBody(row->par);
      

   for (pos = row->pos; pos < cursor.pos; pos++)  {
      if (pos && pos == main_body
	  && !IsLineSeparator(row->par, pos - 1)) {
	     x += LyXStringWidth(style(parameters->textclass, row->par->GetLayout())->labelfont,
				 style(parameters->textclass, row->par->GetLayout())->labelsep);
	     if (x < left_border)
	       x = left_border;
	  }
      
      
      x += SingleWidth(row->par, pos);
      if (HfillExpansion(row, pos)) {
	 if (pos >= main_body)
	   x += fill_hfill;
	 else 
	   x += fill_label_hfill;
      }
      else if (pos >= main_body && IsSeparator(row->par, pos)) {
	 x+= fill_separator;
      }
      
      if (pos + 1 == main_body
	  && IsLineSeparator(row->par, pos)) {
	     x += LyXStringWidth(style(parameters->textclass, row->par->GetLayout())->labelfont,
				 style(parameters->textclass, row->par->GetLayout())->labelsep);
	     if (IsLineSeparator(row->par, pos))
	       x-= SingleWidth(row->par, pos);
	     if (x < left_border)
	       x = left_border;
	  }
   }
   
   cursor.x = (int)x;
   
   
   cursor.x_fix = cursor.x;
   cursor.row = row;
   
   if (cursor.pos && 
       (cursor.pos == cursor.par->Last() || IsSeparator(cursor.par, cursor.pos)
	|| (cursor.pos && cursor.pos == BeginningOfMainBody(cursor.par)
	    && !IsSeparator(cursor.par, cursor.pos))
	)){
	   current_font = cursor.par->GetFontSettings(cursor.pos - 1);
	   real_current_font = GetFont(cursor.par, cursor.pos - 1);
	}
   else {
      current_font = cursor.par->GetFontSettings(cursor.pos);
      real_current_font = GetFont(cursor.par, cursor.pos);
   }
   DeleteEmptyParagraphMechanism(old_cursor);
}

void LyXText::SetCursorFromCoordinates(int x, long y) {
   Row *row;
   int column;
   
   LyXCursor old_cursor;
   
   old_cursor = cursor;
   
   /* get the row first */ 
   
   row = GetRowNearY(y);
   
   cursor.par = row->par;
   
   column = GetColumnNearX(row, x);
   cursor.pos = row->pos + column;
   cursor.x = x;
   cursor.y = y + row->baseline;
   
   cursor.row = row;
    
   if (cursor.pos && 
       (cursor.pos == cursor.par->Last() || IsSeparator(cursor.par, cursor.pos)
	|| (cursor.pos && cursor.pos == BeginningOfMainBody(cursor.par)
	    && !IsSeparator(cursor.par, cursor.pos))
	)){
	   current_font = cursor.par->GetFontSettings(cursor.pos - 1);
	   real_current_font = GetFont(cursor.par, cursor.pos - 1);
	}
   else {
      current_font = cursor.par->GetFontSettings(cursor.pos);
      real_current_font = GetFont(cursor.par, cursor.pos);
   }
   DeleteEmptyParagraphMechanism(old_cursor);
   
}
      

void LyXText::CursorRightOneWord() {
   while (cursor.pos < cursor.par->Last()
	  && !IsSeparator(cursor.par,cursor.pos)
	  && !IsKomma(cursor.par, cursor.pos))
      cursor.pos++;
     while ((cursor.par->Last() > cursor.pos
	     && (IsSeparator(cursor.par,cursor.pos)
		 || IsKomma(cursor.par, cursor.pos)))
	    || (cursor.par->Last() == cursor.pos
		&& cursor.par->Next())){
       if (cursor.pos == cursor.par->Last()) {
	 cursor.par = cursor.par->Next();
	 cursor.pos = 0;
       }
       else
	 cursor.pos++;
     }
     
     SetCursor(cursor.par, cursor.pos);
}

void LyXText::CursorLeftOneWord() {
   while (cursor.pos > 0 
	  && !IsSeparator(cursor.par,cursor.pos-1)
	  && !IsKomma(cursor.par, cursor.pos-1))
     cursor.pos--;

   while (cursor.pos && (IsSeparator(cursor.par, cursor.pos - 1) 
			 || IsKomma(cursor.par, cursor.pos - 1)))
      cursor.pos--;

   if (!cursor.pos) {
     if (cursor.par->Previous()){
       cursor.par = cursor.par->Previous();
       cursor.pos = cursor.par->Last();
     }
     while (cursor.pos && (IsSeparator(cursor.par, cursor.pos - 1) 
			    || IsKomma(cursor.par, cursor.pos - 1)))
       cursor.pos--;
   }
   SetCursor(cursor.par, cursor.pos);
}


void LyXText::SelectWord() {
   /* move cursor to the beginning */
   CursorLeftOneWord();
   if (cursor.par->previous || cursor.pos)
     CursorRightOneWord();
   
   /* set the sel cursor */
   sel_cursor = cursor;
   
   /* now find the end of the word */
   CursorRightOneWord();
   if (cursor.par->next || cursor.pos < cursor.par->Last())
     CursorLeftOneWord();
   
   /* finally set the selection */ 
   SetSelection();
}

char* LyXText::SelectNextWord(float &value) {
   /* move cursor to the beginning */
   
  LyXParagraph* tmppar = cursor.par;


   /* don't forget the very first word */ 
  if (cursor.pos || cursor.par->previous) {
     while (cursor.pos < cursor.par->Last()
	  && !IsSeparator(cursor.par,cursor.pos)
	  && !IsKomma(cursor.par, cursor.pos))
      cursor.pos++;
     while ((cursor.par->Last() > cursor.pos
	     && (IsSeparator(cursor.par,cursor.pos)
		 || IsKomma(cursor.par, cursor.pos)))
	    || (cursor.par->Last() == cursor.pos
		&& cursor.par->Next())){
       if (cursor.pos == cursor.par->Last()) {
	 cursor.par = cursor.par->Next();
	 cursor.pos = 0;
       }
       else
	 cursor.pos++;
     }
  
  }
  
  /* update the value every paragraph */
  if (cursor.par != tmppar){
    SetCursor(cursor.par, cursor.pos);
    value = (float)(cursor.y)/height;
  }

    
   
   /* set the sel cursor */
   sel_cursor = cursor;
   
   /* now find the end of the word */
   while (cursor.pos < cursor.par->Last()
	  && !IsSeparator(cursor.par,cursor.pos)
	  && !IsKomma(cursor.par, cursor.pos))
     cursor.pos++;
   
   char* string = NULL;

   if (sel_cursor.pos < cursor.pos){
     string = new char [cursor.pos - sel_cursor.pos + 5];
     int i;
     for (i=sel_cursor.pos; i<cursor.pos; i++)
       string[i-sel_cursor.pos] = cursor.par->GetChar(i);
     string[i-sel_cursor.pos] = '\0';
   }
   return string;
}
   


void LyXText::SelectSelectedWord(){
  /* move cursor to the beginning */
  SetCursor(sel_cursor.par, sel_cursor.pos);
  
  /* set the sel cursor */
  sel_cursor = cursor;
   
   /* now find the end of the word */
   while (cursor.pos < cursor.par->Last()
	  && !IsSeparator(cursor.par,cursor.pos)
	  && !IsKomma(cursor.par, cursor.pos))
     cursor.pos++;
   
   SetCursor(cursor.par, cursor.pos);

   /* finally set the selection */ 
   SetSelection();
   
}
   

void  LyXText::Delete(){
   LyXCursor old_cursor = cursor;
   
   /* this is a very easy implementation*/ 
   
   /* just move to the right */ 
   CursorRight();
   
   /* if you had success make a backsapce */ 
   if (old_cursor.par != cursor.par || old_cursor.pos != cursor.pos) {
      Backspace();
   }
   

};
void  LyXText::Backspace(){
   LyXParagraph *tmppar;
   Row *tmprow, *row;
   long y;
   int tmpheight;
   LyXFont tmpfont, realtmpfont, reference_font1, reference_font2;

   
   
   tmpfont = current_font;
   realtmpfont = real_current_font;
   
   reference_font1 = GetFont(cursor.par, cursor.par->Last() - 1);
   reference_font2 = cursor.par->GetFontSettings(cursor.par->Last() - 1);
    
   if (cursor.pos == 0) {
      /* we may paste some paragraphs */
      
      /* is it an empty paragraph? */
      
      if ((cursor.par->Last() == 0
	   || (cursor.par->Last() == 1
	       && IsSeparator(cursor.par,0)))
	  && !(cursor.par->Next() 
	       && cursor.par->Next()->footnoteflag == LYX_OPEN_FOOTNOTE)) {
		  
		  
		  if (cursor.par->previous) {
		     tmppar = cursor.par->previous->FirstPhysicalPar();
		     if (cursor.par->GetLayout() == tmppar->GetLayout()
			 && cursor.par->footnoteflag == tmppar->footnoteflag
			 && cursor.par->GetAlign() == tmppar->GetAlign()) {
			    
			    tmppar->line_bottom = cursor.par->line_bottom;
			    tmppar->added_space_bottom = cursor.par->added_space_bottom;
			    tmppar->fill_bottom = cursor.par->fill_bottom;
			    tmppar->pagebreak_bottom = cursor.par->pagebreak_bottom;
			    
			    
			 }
		     
		     CursorLeft();
		     
		     /* the layout things can change the height of a row ! */ 
		     tmpheight = cursor.row->height;
		     SetHeightOfRow(cursor.row);
		     if (cursor.row->height != tmpheight) {
			refresh_y = cursor.y - cursor.row->baseline;
			refresh_row = cursor.row;
			status = LYX_NEED_MORE_REFRESH;
		     }
		     return;
		  }
	       }
      tmppar = cursor.par;
      tmprow = cursor.row;
      CursorLeft();
       /* Pasting is not allowed, if the paragraphs have different layout.
	* I think it is a real bug of all other word processors to allow
	* it. It confuses the user. Even so with a footnote paragraph and
	* a non-footnote paragraph. I will not allow pasting in this case, 
	* because the user would be confused if the footnote behaves 
	* different wether it is open or closed.
	* 
	* Correction: Pasting is always allowed with standard-layout */
      if (cursor.par != tmppar
	  && (cursor.par->GetLayout() == tmppar->GetLayout()
	      || !tmppar->GetLayout())
	  && cursor.par->footnoteflag == tmppar->footnoteflag
	  && cursor.par->GetAlign() == tmppar->GetAlign()) {
	     
	     cursor.par->PasteParagraph();
	     
	     if (!(cursor.pos &&
		   IsSeparator(cursor.par, cursor.pos - 1)))
	       cursor.par->InsertChar(cursor.pos, ' ');
	     else
	       if (cursor.pos)
	       cursor.pos--;
	     
	     
	     status = LYX_NEED_MORE_REFRESH;
	     refresh_row = cursor.row;
	     refresh_y = cursor.y - cursor.row->baseline;
	     
	     /* remove the lost paragraph */
	     RemoveParagraph(tmprow);
	     RemoveRow(tmprow);  
	     
	     /* break the paragraph again */ 
	     /* BreakAgain(cursor.row); */ 
	     
	     AppendParagraph(cursor.row);
	     UpdateCounters(cursor.row);
	     
	     /* the row may have changed, block, hfills etc. */ 
	     SetCursor(cursor.par, cursor.pos);
	 
   
	  }
   }
   else  {
      /* this is the code for a normal backspace, not pasting
       * any paragraphs */ 
      CursorLeft();

      /* some insets are undeletable here */
      if (cursor.par->GetChar(cursor.pos)==LYX_META_INSET) {
	if (!cursor.par->GetInset(cursor.pos)->Deletable())
	  return;
      }
      
      
      row = cursor.row;
      y = cursor.y - row->baseline;
      int z;
     
      /* remember that a space at the end of a row doesnt count
       * when calculating the fill */ 
      if (cursor.pos < RowLast(row) ||
	  !IsLineSeparator(cursor.par,cursor.pos)) {
	     row->fill += SingleWidth(cursor.par, cursor.pos);
	  }
      
      /* some special code when deleting a newline. This is similar
       * to the behavior when pasting paragraphs */ 
      if (cursor.pos && IsNewline(cursor.par, cursor.pos)) {
	 cursor.par->Erase(cursor.pos);
	 /* refresh the positions */
	 tmprow = row;
	 while (tmprow->next && tmprow->next->par == row->par) {
	    tmprow = tmprow->next;
	    tmprow->pos--;
	 }
	 if (IsLineSeparator(cursor.par, cursor.pos - 1))
	    cursor.pos--;
	 
	 if (cursor.pos < cursor.par->Last() && !IsSeparator(cursor.par, cursor.pos)) {
	    cursor.par->InsertChar(cursor.pos, ' ');
	    /* refresh the positions */
	    tmprow = row;
	    while (tmprow->next && tmprow->next->par == row->par) {
	       tmprow = tmprow->next;
	       tmprow->pos++;
	    }
	 }
	 
      }
      else {
	 cursor.par->Erase(cursor.pos);
	 
	 /* refresh the positions */
	 tmprow = row;
	 while (tmprow->next && tmprow->next->par == row->par) {
	    tmprow = tmprow->next;
	    tmprow->pos--;
	 }
	 
	 
	 /* delete superfluous blanks */ 
	 if (cursor.pos < cursor.par->Last() - 1 &&
	     (IsLineSeparator(cursor.par, cursor.pos))) {

		if (cursor.pos == BeginningOfMainBody(cursor.par)
		    || !cursor.pos 
		    || IsLineSeparator(cursor.par, cursor.pos - 1)) {
			   cursor.par->Erase(cursor.pos);
			   /* refresh the positions */
			   tmprow = row;
			   while (tmprow->next && 
				  tmprow->next->par == row->par) {
				     tmprow = tmprow->next;
				     tmprow->pos--;
				  }
			   if (cursor.pos)   /* move one character left */
			     cursor.pos--;
			}
	     }
	 
	 /* delete newlines at the beginning of paragraphs */ 
	 while (cursor.par->Last() &&
		IsNewline(cursor.par, cursor.pos) &&
		cursor.pos == BeginningOfMainBody(cursor.par)
		) {
		   cursor.par->Erase(cursor.pos);
		   /* refresh the positions */
		   tmprow = row;
		   while (tmprow->next && 
			  tmprow->next->par == row->par) {
			     tmprow = tmprow->next;
			     tmprow->pos--;
			  }
		}
	 
      }
			   
      
      /* is there a break one row above */ 
      if (row->previous && row->previous->par == row->par) {
	 z = NextBreakPoint(row->previous, paperwidth);
	 if ( z >= row->pos) {
	    row->pos = z + 1;
	    
	    tmprow = row->previous;
	    
	    /* maybe the current row is now empty */ 
	    if (row->pos >= row->par->Last()) {
	       /* remove it */ 
	       RemoveRow(row);
	       need_break_row = NULL;
	    }
	    else  {
	       BreakAgainOneRow(row);
	       if (row->next && row->next->par == row->par)
	       	 need_break_row = row->next;
	       else
	       	 need_break_row = NULL;
	    }
	    
	    /* set the dimensions of the row above  */ 
	    y -= tmprow->height;
	    tmprow->fill = Fill(tmprow, paperwidth);
	    SetHeightOfRow(tmprow);
	    	    
	    refresh_y = y;
	    refresh_row = tmprow;
	    status = LYX_NEED_MORE_REFRESH;
	    SetCursor(cursor.par, cursor.pos);
	    current_font = tmpfont; real_current_font = realtmpfont;
	    /* check, wether the last characters font has changed. */
	    tmpfont = GetFont(cursor.par, cursor.par->Last() - 1);
	    if (!(FontEqual(reference_font1, tmpfont)
		  || FontEqual(reference_font2, tmpfont)))
	      RedoHeightOfParagraph(cursor);
	    return;
	 }
      }
      
      /* break the cursor row again */ 
      z = NextBreakPoint(row, paperwidth);
      
      
      if ( z != RowLast(row) || 
	  (row->next && row->next->par == row->par &&
	   RowLast(row) == row->par->Last() - 1)){
	      
	      /* it can happen that a paragraph loses one row
	       * without a real breakup. This is when a word
	       * is to long to be broken. Well, I don t care this 
	       * hack ;-) */ 
	      if (row->next && row->next->par == row->par &&
		  RowLast(row) == row->par->Last() - 1)
	      	RemoveRow(row->next);
	      
	      refresh_y = y;
	      refresh_row = row;
	      status = LYX_NEED_MORE_REFRESH;
	      
	      BreakAgainOneRow(row);
	      
	      SetCursor(cursor.par, cursor.pos);
	      /* cursor MUST be in row now */
	      
	      if (row->next && row->next->par == row->par)
	      	need_break_row = row->next;
	      else
	      	need_break_row = NULL;
	   }
      else  {
	 /* set the dimensions of the row */ 
	 row->fill = Fill(row, paperwidth);
	 int tmpheight = row->height;
	 SetHeightOfRow(row);
	 if (tmpheight == row->height)
	   status = LYX_NEED_VERY_LITTLE_REFRESH;
	 else
	   status = LYX_NEED_MORE_REFRESH;
   	 refresh_y = y;
	 refresh_row = row;
	 SetCursor(cursor.par, cursor.pos);
      }
   }
   
   /* restore the current font 
    * That is what a user
    * expects! */
   current_font = tmpfont; real_current_font = realtmpfont;

   /* check, wether the last characters font has changed. */
   tmpfont = GetFont(cursor.par, cursor.par->Last() - 1);
   if (!(FontEqual(reference_font1, tmpfont)
	 || FontEqual(reference_font2, tmpfont)))
     RedoHeightOfParagraph(cursor);
   else {
      /* now the special right address boxes */
      if (style(parameters->textclass, cursor.par->GetLayout())->bordertype == BORDER_RIGHT_ADDRESS_BOX) {
	 RedoDrawingOfParagraph(cursor); 
      }
   }

   
}
   
   
   







unsigned long LyXText::GetVisibleRow(Row *row_ptr, long y) {
   /* returns a printed row */
   
   int screendepth;
   int pos, pos_end, lg;
   float x, tmpx;
   Pixmap pixmap;
   int y2;
   float fill_separator, fill_hfill, fill_label_hfill;
   LyXParagraph *par, *firstpar;
   int left_border;
   LyXFont font;
   char* tmpstring;
   int maxasc;
   int maxdesc;
   if (row_ptr->height <= 0) {
      fprintf(stderr, "LYX_ERROR: row.height: %d \n", row_ptr->height);
      return 0;
   }
   left_border = LabelEnd(row_ptr);
   PrepareToPrint(row_ptr, x, fill_separator, fill_hfill, fill_label_hfill);

   int main_body = BeginningOfMainBody(row_ptr->par);
   
   
   /* create and initialize the pixmap */
   
   screendepth= fl_get_visual_depth();
   pixmap = XCreatePixmap (fl_display,
			   RootWindow(fl_display, 0),
			   paperwidth, row_ptr->height, screendepth);

   XFillRectangle(fl_display,pixmap,
		  LyXGetClearGC(),
		  0, 0, paperwidth, row_ptr->height);
   
   
   if (row_ptr->par->footnoteflag == LYX_OPEN_FOOTNOTE) {
      if (row_ptr->previous && 
	  row_ptr->previous->par->footnoteflag != LYX_OPEN_FOOTNOTE) {
	     XDrawLine(fl_display,pixmap,LyXGetCopyGC(),
		       0, 0, paperwidth, 0);
	  }
      
/*      if (row_ptr->next && 
	  row_ptr->next->par->footnoteflag != LYX_OPEN_FOOTNOTE) {
	     XDrawLine(fl_display,pixmap,LyXGetCopyGC(),
		       0, row_ptr->height - 1, 
		       paperwidth, row_ptr->height - 1);
	  }
*/
	     
   }
   else  {
      if (row_ptr->previous && 
	  row_ptr->previous->par->footnoteflag == LYX_OPEN_FOOTNOTE) {
	     XDrawLine(fl_display,pixmap,LyXGetCopyGC(),
		       0, 0, paperwidth, 0);
	  }
   }
   
   
   LyXLayout* layout = style(parameters->textclass, row_ptr->par->GetLayout());
   firstpar = row_ptr->par->FirstPhysicalPar();
   
   /* is it a first row? */ 
    if (row_ptr->pos == 0
	&& row_ptr->par == firstpar) {
	   
	   /* should we print a label? */ 
	   if (layout->labeltype >= LABEL_STATIC) {
	      font = layout->labelfont;
	      if (row_ptr->par->GetLabelString()) {
		 tmpx = x;
		 tmpstring = row_ptr->par->GetLabelString();
		 for (lg=0;tmpstring[lg];lg++);
		 
		 if (layout->labeltype == LABEL_COUNTER_CHAPTER) {
		    if (parameters->secnumdepth >=0){
		       /* this is spacel code for the chapter layout. This is printed in
			* an extra row and has a pagebreak at the top. */ 
		       maxdesc = (int) (LyXMaxDescent(font) * layout->baselinestretch * parameters->baseline);
		       maxasc = (int) (LyXMaxAscent(font) * layout->baselinestretch * parameters->baseline);
		       XDrawString(fl_display,
				   pixmap,
				   LyXGetGC(font),
				   (int) x, row_ptr->baseline - maxdesc - maxasc - (int) layout->bottomsep * DefaultHeight(),
				   tmpstring, lg);
		       /* DrawOnOffLine(pixmap, 
					2 * DefaultHeight(),
					0, paperwidth); */ 
		    }
		 }
		 else {
		    x -= LyXStringWidth(font, layout->labelsep);
		    x -= LyXStringWidth(font, tmpstring);
		    /* draw it! */
		    XDrawString(fl_display,
				pixmap,
				LyXGetGC(font),
				(int) x, row_ptr->baseline,
				tmpstring, lg);
		 }
		 
		 x = tmpx;
	      }
	   }
       /* the labels at the top of an environment. More or less for bibliography */ 
       else if (layout->labeltype == LABEL_TOP_ENVIRONMENT) {
	  if (!row_ptr->par->Previous() 
	      || row_ptr->par->Previous()->GetLayout() != row_ptr->par->GetLayout()) {
	     LyXLayout *top_label_layout = style(parameters->textclass, NumberOfLayout(parameters->textclass, 
										  TextClass(parameters->textclass)->top_label_layout));
	     font = top_label_layout->labelfont;
	     
	     if (row_ptr->par->GetLabelString()) {
		tmpstring = row_ptr->par->GetLabelString();
		for (lg=0;tmpstring[lg];lg++);
		
		maxasc = (int) (LyXMaxAscent(top_label_layout->labelfont) * top_label_layout->baselinestretch * parameters->baseline)
		+ (int) top_label_layout->topsep * DefaultHeight();
		
		XDrawString(fl_display,
			    pixmap,
			    LyXGetGC(font),
			    (int) x, maxasc,
			    tmpstring, lg);
	     }
	  }
       }
       
       
       if (row_ptr->par->line_top) {      /* draw a top line  */
	  y2 =  (LyXAscent(GetFont(row_ptr->par, 0),'x')
		    + int (row_ptr->par->added_space_top * 37)
		    + row_ptr->par->pagebreak_top * 3 * DefaultHeight()
		    + row_ptr->par->fill_top * 3 * DefaultHeight());
	      
	      /* think about the parskip */ 
	      	  
	      /* some parksips VERY EASY IMPLEMENTATION */ 
	      if (parameters->paragraph_separation == LYX_PARSEP_SKIP) {
		 if (layout->latextype == LATEX_PARAGRAPH
		     && firstpar->GetDepth() == 0
		     && firstpar->Previous())
		   y2 += (int)(layout->parskip * DefaultHeight());
		 else if (firstpar->Previous()
			  && style(parameters->textclass, firstpar->Previous()->GetLayout())->latextype == LATEX_PARAGRAPH
			  && firstpar->Previous()->GetDepth() == 0)
		   y2 += (int)(style(parameters->textclass, firstpar->Previous()->GetLayout())->parskip * DefaultHeight());
		 
	      }
	      
	      
	      /* think about the borders */ 
	      if (!row_ptr->previous)
	      	y2 += LYX_PAPER_MARGIN;
	      
	      DrawThickLine(pixmap,
			    y2,
			    0, paperwidth);
	   }
	   
	   if (row_ptr->par->pagebreak_top) /* draw a top pagebreak  */
	   DrawOnOffLine(pixmap, 
			 2 * DefaultHeight(),
			 0, paperwidth);
	   
	   if (row_ptr->par->fill_top) {    /* draw a vfill top  */
	      DrawLine(pixmap, 
		       2 + row_ptr->par->pagebreak_top * 3 * DefaultHeight(), 
		       0, LYX_PAPER_MARGIN);
	      DrawLine(pixmap, 
		       3 * DefaultHeight()
		       + row_ptr->par->pagebreak_top * 3 * DefaultHeight(),
		       0, LYX_PAPER_MARGIN);
	      DrawVerticalOnOffLine(pixmap, LYX_PAPER_MARGIN / 2, 
				    2 + row_ptr->par->pagebreak_top * 3 * DefaultHeight(),
				    3 * DefaultHeight()
				    + row_ptr->par->pagebreak_top * 3 * DefaultHeight()
				    );
	   }
	}
   
   /* is it a last row? */
   
   par = row_ptr->par->LastPhysicalPar();
   if (row_ptr->par->ParFromPos(RowLast(row_ptr) + 1) == par
       && (!row_ptr->next || row_ptr->next->par != row_ptr->par)) {     
	  if (firstpar->line_bottom) {
	     /* draw a bottom line */
	     y2 = (row_ptr->height - 
		   LyXAscent(GetFont(par, par->Last() - 1),'x')
		   - int (firstpar->added_space_bottom * 37)
		   - firstpar->pagebreak_bottom * 3 * DefaultHeight()
		   - firstpar->fill_bottom * 3 * DefaultHeight());
	     
	     /* think about the borders */ 
	     if (!row_ptr->next)
	       y2 -= LYX_PAPER_MARGIN;
	     
	     DrawThickLine(pixmap,
			   y2,
			   0, paperwidth);
	  }
	  /* draw a bottom pagebreak */ 
	  if (firstpar->pagebreak_bottom)
	    DrawOnOffLine(pixmap, 
			  row_ptr->height - 2 * DefaultHeight(),
			  0, paperwidth);
	  
	  if (firstpar->fill_bottom) {    /* draw a vfill bottom  */
	     DrawLine(pixmap, 
		      row_ptr->height - 3 * DefaultHeight()
		      - firstpar->pagebreak_bottom * 3 * DefaultHeight(),
		      0, LYX_PAPER_MARGIN);
	     DrawLine(pixmap, row_ptr->height 
		      - 2 -  firstpar->pagebreak_bottom * 3 * DefaultHeight(), 
		      0, LYX_PAPER_MARGIN);
	     DrawVerticalOnOffLine(pixmap, LYX_PAPER_MARGIN / 2, 
				   row_ptr->height - 3 * DefaultHeight()
				   - firstpar->pagebreak_bottom * 3 * DefaultHeight(),
				   row_ptr->height
				   - 2 -  firstpar->pagebreak_bottom * 3 * DefaultHeight()
				   );
	  }
       }
   
   
   /* draw the text in the pixmap */  
   pos_end = RowLast(row_ptr);
   
   
   
   pos = row_ptr->pos;
   
   while (pos <= pos_end)  {

      if (IsHfill(row_ptr->par, pos)) {
	 x += 1;
	 DrawVerticalLine(pixmap, (int)x,  
			  row_ptr->baseline - DefaultHeight()/2, 
			  row_ptr->baseline);
	 if (HfillExpansion(row_ptr,pos)) {
	    if (pos >= main_body) {
	       DrawOnOffLine(pixmap,
			     row_ptr->baseline - DefaultHeight()/4,
			     (int)x,
			     (int)fill_hfill);
	       x += fill_hfill;
	    }
	    else {
	       DrawOnOffLine(pixmap,
			     row_ptr->baseline - DefaultHeight()/4,
			     (int)x,
			     (int)fill_label_hfill);
	       x += fill_label_hfill;
	    }
	    DrawVerticalLine(pixmap, (int)x,
			     row_ptr->baseline - DefaultHeight()/2, 
			     row_ptr->baseline);
	 }
	 x += 2;
	 pos++;
      }
      else {
	 if (IsSeparator(row_ptr->par, pos)) {
	    tmpx = x;
	    x+=SingleWidth(row_ptr->par, pos);
	    if (pos >= main_body)
	      x+= fill_separator;
	    if (row_ptr->par->GetChar(pos)==LYX_META_PROTECTED_SEPARATOR) {
	       DrawVerticalLine(pixmap, (int) tmpx,
				row_ptr->baseline - 3,
				row_ptr->baseline - 1);
	       DrawLine(pixmap, 
			row_ptr->baseline - 1,
			(int)tmpx,
			(int)(x-tmpx-2));
	       DrawVerticalLine(pixmap, (int) (x-2),
				row_ptr->baseline - 3,
				row_ptr->baseline - 1);
	       /* what about underbars? */
	       font = GetFont(row_ptr->par, pos); 
	       if (font.bar == LYX_UNDER_BAR
		   && font.latex == LYX_NO_LATEX){
		 XDrawLine(fl_display,
			   pixmap,
			   LyXGetCopyGC(),
			   (int)tmpx, row_ptr->baseline + 2,
			   (int)x, row_ptr->baseline + 2);
	       }
	    }
	    pos++;
	 }
	 else
	    Draw(row_ptr, pos, pixmap, x);
      }
      if (pos == main_body) {
	 x += LyXStringWidth(layout->labelfont,
			     layout->labelsep);
	 if (IsLineSeparator(row_ptr->par, pos - 1))
	   x-= SingleWidth(row_ptr->par, pos - 1);
	 if (x < left_border)
	   x = left_border;
      }
   }
   

   

   if (selection) {
         
      /* selection code */ 
      if (sel_start_cursor.row == row_ptr && sel_end_cursor.row == row_ptr) {
	 XFillRectangle(fl_display,pixmap,
			LyXGetSelectGC(),
			sel_start_cursor.x, 0, 
			sel_end_cursor.x - sel_start_cursor.x, row_ptr->height);
      }
      else if (sel_start_cursor.row == row_ptr) {
	 XFillRectangle(fl_display,pixmap,
			LyXGetSelectGC(),
			sel_start_cursor.x, 0, 
			paperwidth - sel_start_cursor.x, row_ptr->height);
      }
      else if (sel_end_cursor.row == row_ptr) {
	 XFillRectangle(fl_display,pixmap,
			LyXGetSelectGC(),
			0, 0, 
			sel_end_cursor.x, row_ptr->height);
      }
      else if (y > sel_start_cursor.y && y < sel_end_cursor.y) {
	 XFillRectangle(fl_display,pixmap,
			LyXGetSelectGC(),
			0, 0, paperwidth, row_ptr->height);
	 
      }
   }
      
      
    
   /* thats it */ 
   return pixmap;
}



int LyXText::DefaultHeight(){
   LyXLayout *layout = NULL;
   if (!layout)
      layout = style(parameters->textclass, 0);
   
   return (int) ((LyXMaxAscent(layout->font)
   + LyXMaxDescent(layout->font)) * layout->baselinestretch);
}

   

   
   
   
/* returns the column near the specified x-coordinate of the row 
* x is set to the real beginning of this column  */ 
int  LyXText::GetColumnNearX(Row *row, int& x) {
   int c; 
   float tmpx;
   float fill_separator, fill_hfill, fill_label_hfill;
   int left_border;
   
   left_border = LabelEnd(row);
   PrepareToPrint(row, tmpx, fill_separator, fill_hfill, fill_label_hfill);
   int main_body = BeginningOfMainBody(row->par);
   
   c = row->pos;

   int last = RowLast(row);
   if (IsNewline(row->par, last))
     last--;
   
   LyXLayout *layout = style(parameters->textclass, row->par->GetLayout());
   
   while (c <= last
	  && tmpx + SingleWidth(row->par, c)  <= x) {

	     if (c && c == main_body
		 && !IsLineSeparator(row->par, c - 1)) {
		    tmpx += LyXStringWidth(layout->labelfont,
					   layout->labelsep);
		    if (tmpx < left_border)
		      tmpx = left_border;
		 }
	     
	     tmpx += SingleWidth(row->par, c);
	     if (HfillExpansion(row, c)) {
		if (c >= main_body)
		  tmpx += fill_hfill;
		else
		  tmpx += fill_label_hfill;
	     }
	     else if (c >= main_body && IsSeparator(row->par, c)) {
		tmpx+= fill_separator;  
	     }
	     c++;
	     if (c == main_body
		 && IsLineSeparator(row->par, c - 1)) {
		    tmpx += LyXStringWidth(layout->labelfont,
					   layout->labelsep);
		    tmpx-= SingleWidth(row->par, c - 1);
		    if (tmpx < left_border)
		      tmpx = left_border;
		 }
	     
	  }
   /* make shure that a last space in a row doesnt count */
   if (c >= last 
       && IsLineSeparator(row->par, c - 1)
       && !(!row->next || row->next->par != row->par)) {
	  tmpx -= SingleWidth(row->par, c - 1);
	  tmpx -= fill_separator;
       }
   c-=row->pos;
   
   x = (int)tmpx;
   return c;
}

   




/* turn the selection into a new environement. If there is no selection,
* create an empty environement */ 
void LyXText::InsertFootnoteEnvironement(char kind) {
   
   /* no footnoteenvironement in a footnoteenvironement */ 
   if (cursor.par->footnoteflag != LYX_NO_FOOTNOTE) {
      WriteAlert("Impossible operation", "You can't insert a float in a float!", "sorry.");
     return;
   }
   
   /* this doesnt make sense, if there is no selection */ 
   char dummy_selection = False;
   if (!selection) {
      sel_start_cursor = cursor;       /* dummy selection  */
      sel_end_cursor = cursor;
      dummy_selection = True;
   }
   
   
   LyXParagraph *tmppar;

   /* a test to make shure there is not already a footnote
    * in the selection. */
   
   tmppar = sel_start_cursor.par->ParFromPos(sel_start_cursor.pos);
   
   while (tmppar != sel_end_cursor.par->ParFromPos(sel_end_cursor.pos) && 
	  tmppar->footnoteflag == LYX_NO_FOOTNOTE)
     tmppar = tmppar->next;
   
   if (tmppar != sel_end_cursor.par->ParFromPos(sel_end_cursor.pos)
       || tmppar->footnoteflag != LYX_NO_FOOTNOTE) {
      WriteAlert("Impossible operation", "Float  would include float!", "sorry.");
      return;
   }
   
   
   /* ok we have a selection. This is always between sel_start_cursor
    * and sel_end cursor */
   
   
   if (sel_end_cursor.pos > 0 
       && IsLineSeparator(sel_end_cursor.par,sel_end_cursor.pos - 1))
     sel_end_cursor.pos--;	       /* please break before a space at
					* the end */
   if (sel_start_cursor.par == sel_end_cursor.par
       && sel_start_cursor.pos > sel_end_cursor.pos)
     sel_start_cursor.pos--;

   sel_end_cursor.par->BreakParagraphConservative(sel_end_cursor.pos);
   
   sel_end_cursor.par = sel_end_cursor.par->Next();
   sel_end_cursor.pos = 0;
   
   if (sel_start_cursor.par->GetLayout() == sel_end_cursor.par->layout)
     sel_end_cursor.par->layout = LYX_DUMMY_LAYOUT;

   cursor = sel_end_cursor;

   
   /* please break behind a space, if there is one. The space should
    * be erased too */ 
   if (sel_start_cursor.pos > 0 
       && IsLineSeparator(sel_start_cursor.par,sel_start_cursor.pos - 1))
     sel_start_cursor.pos--;
   if (IsLineSeparator(sel_start_cursor.par,sel_start_cursor.pos)) {
      sel_start_cursor.par->Erase(sel_start_cursor.pos);
   }
   
   sel_start_cursor.par->BreakParagraphConservative(sel_start_cursor.pos);
   tmppar = sel_start_cursor.par->Next();
   

   if (dummy_selection) {
     tmppar->Clear();
     if (kind == LYX_TAB
	 || kind == LYX_FIG)
       tmppar->SetLayout(NumberOfLayout(parameters->textclass,"Caption"));
   }
   else {
     if (sel_start_cursor.pos > 0) {
       /* the footnote-environement should beginn with standard layout.
	* Imaging you insert a footnote withing a enumeration, you 
	* certainly do not want a enumerated footnote! */ 
       tmppar->Clear();
     }
     else {
       /* this is a exception the user would sometimes expect, I hope */
       sel_start_cursor.par->Clear();
     }
   }
   
   
   
   while (tmppar != sel_end_cursor.par) {
      tmppar->footnoteflag = LYX_OPEN_FOOTNOTE;
      tmppar->footnotekind = kind;
      tmppar = tmppar->Next();
   } 

   RedoParagraphs(sel_start_cursor, sel_end_cursor.par->Next());
   
   SetCursor(sel_start_cursor.par->Next(), 0);

   ClearSelection();
}
   



/* returns pointer to a specified row */
Row* LyXText::GetRow(LyXParagraph *par, int pos, long &y) {
   Row* tmprow;
   LyXParagraph *tmppar;
   if (currentrow){
     if (par == currentrow->par || par == currentrow->par->Previous()){
       while (currentrow->previous &&
	      currentrow->previous->par != par->previous){
	 currentrow = currentrow->previous;
	 currentrow_y -= currentrow->height;
       }
     }
     tmprow = currentrow;
     y = currentrow_y;
     /* find the first row of the specified paragraph */ 
     while (tmprow->next && (tmprow->par != par)) {
       y += tmprow->height;
       tmprow = tmprow->next;
     }

     if (tmprow->par == par){
       /* now find the wanted row */ 
       while (tmprow->pos < pos && tmprow->next && tmprow->next->par == par && 
	      tmprow->next->pos <= pos) {
	 y += tmprow->height;
       tmprow = tmprow->next;
       }
       currentrow = tmprow;
       currentrow_y = y;
       return tmprow;
     }
   }
   tmprow = firstrow;
   y = 0;
   /* find the first row of the specified paragraph */ 
   while (tmprow->next && (tmprow->par != par)) {
      y += tmprow->height;
      tmprow = tmprow->next;
   }
 
   /* now find the wanted row */ 
   while (tmprow->pos < pos && tmprow->next && tmprow->next->par == par && 
	  tmprow->next->pos <= pos) {
	     y += tmprow->height;
	     tmprow = tmprow->next;
   }
   
   currentrow = tmprow;
   currentrow_y = y;

   return tmprow;
}


	  


